# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


import wrapt
import re

from typing import Any, Callable, Dict, Tuple, Union
from instana.log import logger
from instana.instrumentation.google.cloud.collectors import _storage_api
from instana.util.traceutils import get_tracer_tuple, tracing_is_off

try:
    from google.cloud import storage

    logger.debug("Instrumenting google-cloud-storage")

    def _collect_attributes(
        api_request: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Extract span tags from Google Cloud Storage API request. Returns None if the request is not
        supported.

        :param: dict
        :return: dict or None
        """
        method, path = api_request.get("method", None), api_request.get("path", None)

        if method not in _storage_api:
            return

        try:
            params = api_request.get("query_params", {})
            data = api_request.get("data", {})

            if path in _storage_api[method]:
                # check is any of string keys matches the path exactly
                return _storage_api[method][path](params, data)
            else:
                # look for a regex that matches the string
                for matcher, collect in _storage_api[method].items():
                    if not isinstance(matcher, re.Pattern):
                        continue

                    m = matcher.match(path)
                    if m is None:
                        continue

                    return collect(params, data, m)
        except Exception:
            logger.debug(
                "instana.instrumentation.google.cloud.storage._collect_attributes: ",
                exc_info=True,
            )

    def execute_with_instana(
        wrapped: Callable[..., object],
        instance: Union[storage.Batch, storage._http.Connection],
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        # batch requests are traced with finish_batch_with_instana()
        # also return early if we're not tracing
        if isinstance(instance, storage.Batch) or tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span("gcs", span_context=parent_context) as span:
            try:
                attributes = _collect_attributes(kwargs)

                # don't trace if the call is not instrumented
                if attributes is None:
                    logger.debug(
                        f"uninstrumented Google Cloud Storage API request: {kwargs}"
                    )
                    return wrapped(*args, **kwargs)
                span.set_attributes(attributes)
                kv = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return kv

    def download_with_instana(
        wrapped: Callable[..., object],
        instance: storage.Blob,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        # return early if we're not tracing
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span("gcs", span_context=parent_context) as span:
            span.set_attribute("gcs.op", "objects.get")
            span.set_attribute("gcs.bucket", instance.bucket.name)
            span.set_attribute("gcs.object", instance.name)

            start = len(args) > 4 and args[4] or kwargs.get("start", None)
            if start is None:
                start = ""

            end = len(args) > 5 and args[5] or kwargs.get("end", None)
            if end is None:
                end = ""

            if start != "" or end != "":
                span.set_attribute("gcs.range", "-".join((start, end)))

            try:
                kv = wrapped(*args, **kwargs)
            except Exception as e:
                span.record_exception(e)
            else:
                return kv

    def upload_with_instana(
        wrapped: Callable[..., object],
        instance: storage.Blob,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        # return early if we're not tracing
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span("gcs", span_context=parent_context) as span:
            span.set_attribute("gcs.op", "objects.insert")
            span.set_attribute("gcs.bucket", instance.bucket.name)
            span.set_attribute("gcs.object", instance.name)

            try:
                kv = wrapped(*args, **kwargs)
            except Exception as e:
                span.record_exception(e)
            else:
                return kv

    def finish_batch_with_instana(
        wrapped: Callable[..., object],
        instance: storage.Batch,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        # return early if we're not tracing
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span("gcs", span_context=parent_context) as span:
            span.set_attribute("gcs.op", "batch")
            span.set_attribute("gcs.projectId", instance._client.project)
            span.set_attribute("gcs.numberOfOperations", len(instance._requests))

            try:
                kv = wrapped(*args, **kwargs)
            except Exception as e:
                span.record_exception(e)
            else:
                return kv

    wrapt.wrap_function_wrapper(
        "google.cloud.storage._http", "Connection.api_request", execute_with_instana
    )
    wrapt.wrap_function_wrapper(
        "google.cloud.storage.blob", "Blob._do_download", download_with_instana
    )
    wrapt.wrap_function_wrapper(
        "google.cloud.storage.blob", "Blob._do_upload", upload_with_instana
    )
    wrapt.wrap_function_wrapper(
        "google.cloud.storage.batch", "Batch.finish", finish_batch_with_instana
    )
except ImportError:
    pass



from multiprocessing import Queue
import sys
from unittest import TestCase
from unittest.mock import NonCallableMagicMock, PropertyMock

import pytest

from instana.recorder import StanRecorder


@pytest.mark.skipif(
    sys.platform == "darwin",
    reason="Avoiding NotImplementedError when calling multiprocessing.Queue.qsize()",
)
class TestStanRecorderTC(TestCase):
    def setUp(self):
        mock_agent = NonCallableMagicMock()
        mock_collector = NonCallableMagicMock(span_queue=Queue())
        mock_agent.collector = mock_collector
        self.recorder = StanRecorder(agent=mock_agent)
        self.mock_suppressed_span = NonCallableMagicMock()
        self.mock_suppressed_span.context = NonCallableMagicMock()
        self.mock_suppressed_property = PropertyMock(return_value=True)
        type(
            self.mock_suppressed_span.context
        ).suppression = self.mock_suppressed_property

    def test_record_span_with_suppression(self):
        # Ensure that the queue is empty
        self.assertEqual(self.recorder.queue_size(), 0)
        self.recorder.record_span(self.mock_suppressed_span)
        # Ensure that even after adding a suppressed span
        # the queue remains empty
        self.assertEqual(self.recorder.queue_size(), 0)
        # Ensure that the no recorded spans can be retrieved
        self.assertEqual(self.recorder.queued_spans(), [])

        # Make sure that the success so far has indeed resulted after a getitem
        # call to the 'suppression' property of the mock span context
        self.mock_suppressed_property.assert_called_once_with()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import os
import threading
import time
from typing import Generator

import pytest
import six
from google.api_core.exceptions import AlreadyExists
from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient
from google.cloud.pubsub_v1.publisher import exceptions
from opentelemetry.trace import SpanKind

from instana.singletons import agent, tracer
from instana.span.span import get_current_span
from tests.test_utils import _TraceContextMixin

# Use PubSub Emulator exposed at :8085
os.environ["PUBSUB_EMULATOR_HOST"] = "localhost:8681"


class TestPubSubPublish(_TraceContextMixin):
    publisher = PublisherClient()

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        self.project_id = "test-project"
        self.topic_name = "test-topic"

        # setup topic_path & topic
        self.topic_path = self.publisher.topic_path(self.project_id, self.topic_name)
        try:
            self.publisher.create_topic(request={"name": self.topic_path})
        except AlreadyExists:
            self.publisher.delete_topic(request={"topic": self.topic_path})
            self.publisher.create_topic(request={"name": self.topic_path})
        yield
        self.publisher.delete_topic(request={"topic": self.topic_path})
        agent.options.allow_exit_as_root = False

    def test_publish(self) -> None:
        # publish a single message
        with tracer.start_as_current_span("test"):
            future = self.publisher.publish(
                self.topic_path, b"Test Message", origin="instana"
            )
        time.sleep(2.0)  # for sanity
        result = future.result()
        assert isinstance(result, six.string_types)

        spans = self.recorder.queued_spans()
        gcps_span, test_span = spans[0], spans[1]

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()
        assert gcps_span.n == "gcps"
        assert gcps_span.k is SpanKind.CLIENT

        assert gcps_span.data["gcps"]["op"] == "publish"
        assert self.topic_name == gcps_span.data["gcps"]["top"]

        # Trace Context Propagation
        self.assertTraceContextPropagated(test_span, gcps_span)

        # Error logging
        self.assertErrorLogging(spans)

    def test_publish_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        # publish a single message
        future = self.publisher.publish(
            self.topic_path, b"Test Message", origin="instana"
        )
        time.sleep(2.0)  # for sanity
        result = future.result()
        assert isinstance(result, six.string_types)

        spans = self.recorder.queued_spans()
        assert len(spans) == 1
        gcps_span = spans[0]

        current_span = get_current_span()
        assert not current_span.is_recording()
        assert gcps_span.n == "gcps"
        assert gcps_span.k is SpanKind.CLIENT

        assert gcps_span.data["gcps"]["op"] == "publish"
        assert self.topic_name == gcps_span.data["gcps"]["top"]

        # Error logging
        self.assertErrorLogging(spans)


class AckCallback(object):
    def __init__(self) -> None:
        self.calls = 0
        self.lock = threading.Lock()

    def __call__(self, message) -> None:
        message.ack()
        # Only increment the number of calls **after** finishing.
        with self.lock:
            self.calls += 1


class TestPubSubSubscribe(_TraceContextMixin):
    @classmethod
    def setup_class(cls) -> None:
        cls.publisher = PublisherClient()
        cls.subscriber = SubscriberClient()

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        self.project_id = "test-project"
        self.topic_name = "test-topic"
        self.subscription_name = "test-subscription"

        # setup topic_path & topic
        self.topic_path = self.publisher.topic_path(self.project_id, self.topic_name)
        try:
            self.publisher.create_topic(request={"name": self.topic_path})
        except AlreadyExists:
            self.publisher.delete_topic(request={"topic": self.topic_path})
            self.publisher.create_topic(request={"name": self.topic_path})

        # setup subscription path & attach subscription
        self.subscription_path = self.subscriber.subscription_path(
            self.project_id,
            self.subscription_name,
        )
        try:
            self.subscriber.create_subscription(
                request={"name": self.subscription_path, "topic": self.topic_path}
            )
        except AlreadyExists:
            self.subscriber.delete_subscription(
                request={"subscription": self.subscription_path}
            )
            self.subscriber.create_subscription(
                request={"name": self.subscription_path, "topic": self.topic_path}
            )
        yield
        self.publisher.delete_topic(request={"topic": self.topic_path})
        self.subscriber.delete_subscription(
            request={"subscription": self.subscription_path}
        )

    def test_subscribe(self) -> None:
        with tracer.start_as_current_span("test"):
            # Publish a message
            future = self.publisher.publish(
                self.topic_path, b"Test Message to PubSub", origin="instana"
            )
            assert isinstance(future.result(), six.string_types)

            time.sleep(2.0)  # for sanity

            # Subscribe to the subscription
            callback_handler = AckCallback()
            future = self.subscriber.subscribe(self.subscription_path, callback_handler)
            timeout = 2.0
            try:
                future.result(timeout)
            except exceptions.TimeoutError:
                future.cancel()

        spans = self.recorder.queued_spans()

        producer_span = spans[0]
        consumer_span = spans[1]
        test_span = spans[2]

        assert len(spans) == 3
        current_span = get_current_span()
        assert not current_span.is_recording()
        assert producer_span.data["gcps"]["op"] == "publish"
        assert consumer_span.data["gcps"]["op"] == "consume"
        assert self.topic_name == producer_span.data["gcps"]["top"]
        assert self.subscription_name == consumer_span.data["gcps"]["sub"]

        assert producer_span.k is SpanKind.CLIENT
        assert consumer_span.k is SpanKind.SERVER

        # Trace Context Propagation
        self.assertTraceContextPropagated(producer_span, consumer_span)
        self.assertTraceContextPropagated(test_span, producer_span)

        # Error logging
        self.assertErrorLogging(spans)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import logging
import pytest

from typing import Generator
from instana.instrumentation.psycopg2 import register_json_with_instana
from tests.helpers import testenv
from instana.singletons import agent, tracer

import psycopg2
import psycopg2.extras
import psycopg2.extensions as ext

logger = logging.getLogger(__name__)


class TestPsycoPG2:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        kwargs = {
            "host": testenv["postgresql_host"],
            "port": testenv["postgresql_port"],
            "user": testenv["postgresql_user"],
            "password": testenv["postgresql_pw"],
            "dbname": testenv["postgresql_db"],
        }
        self.db = psycopg2.connect(**kwargs)

        database_setup_query = """
        DROP TABLE IF EXISTS users;
        CREATE TABLE users(
           id serial PRIMARY KEY,
           name VARCHAR (50),
           password VARCHAR (50),
           email VARCHAR (355),
           created_on TIMESTAMP,
           last_login TIMESTAMP
        );
        INSERT INTO users(name, email) VALUES('kermit', 'kermit@muppets.com');
        DROP FUNCTION IF EXISTS test_proc(VARCHAR(70));
        CREATE FUNCTION test_proc(candidate VARCHAR(70))
        RETURNS text AS $$
        BEGIN
            RETURN(SELECT name FROM users where email = candidate);
        END;
        $$ LANGUAGE plpgsql;
        """
        cursor = self.db.cursor()
        cursor.execute(database_setup_query)
        self.db.commit()

        self.cursor = self.db.cursor()
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        tracer.cur_ctx = None
        yield
        if self.cursor and not self.cursor.connection.closed:
            self.cursor.close()
        if self.db and not self.db.closed:
            self.db.close()
        agent.options.allow_exit_as_root = False

    def test_register_json(self) -> None:
        resp = register_json_with_instana(conn_or_curs=self.db)
        assert resp[0].values[0] == 114
        assert resp[1].values[0] == 199

    def test_vanilla_query(self) -> None:
        assert psycopg2.extras.register_uuid(None, self.db)
        assert psycopg2.extras.register_uuid(None, self.db.cursor())

        self.cursor.execute("""SELECT * from users""")
        affected_rows = self.cursor.rowcount
        assert affected_rows == 1
        result = self.cursor.fetchone()

        assert len(result) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 0

    def test_basic_query(self) -> None:
        with tracer.start_as_current_span("test"):
            self.cursor.execute("""SELECT * from users""")
            affected_rows = self.cursor.rowcount
            result = self.cursor.fetchone()
            self.db.commit()

        assert affected_rows == 1
        assert len(result) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert db_span.data["pg"]["stmt"] == "SELECT * from users"
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_basic_query_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.cursor.execute("""SELECT * from users""")
        affected_rows = self.cursor.rowcount
        result = self.cursor.fetchone()
        self.db.commit()

        assert affected_rows == 1
        assert len(result) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        db_span = spans[0]

        assert not db_span.ec

        assert db_span.n, "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert db_span.data["pg"]["stmt"] == "SELECT * from users"
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_basic_insert(self) -> None:
        with tracer.start_as_current_span("test"):
            self.cursor.execute(
                """INSERT INTO users(name, email) VALUES(%s, %s)""",
                ("beaker", "beaker@muppets.com"),
            )
            affected_rows = self.cursor.rowcount

        assert affected_rows == 1

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert (
            db_span.data["pg"]["stmt"]
            == "INSERT INTO users(name, email) VALUES(%s, %s)"
        )
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_executemany(self) -> None:
        with tracer.start_as_current_span("test"):
            self.cursor.executemany(
                "INSERT INTO users(name, email) VALUES(%s, %s)",
                [("beaker", "beaker@muppets.com"), ("beaker", "beaker@muppets.com")],
            )
            affected_rows = self.cursor.rowcount
            self.db.commit()

        assert affected_rows == 2

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert (
            db_span.data["pg"]["stmt"]
            == "INSERT INTO users(name, email) VALUES(%s, %s)"
        )

        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_call_proc(self) -> None:
        with tracer.start_as_current_span("test"):
            callproc_result = self.cursor.callproc("test_proc", ("beaker",))

        assert isinstance(callproc_result, tuple)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert db_span.data["pg"]["stmt"] == "test_proc"
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_error_capture(self) -> None:
        affected_rows = result = None
        try:
            with tracer.start_as_current_span("test"):
                self.cursor.execute("""SELECT * from blah""")
                affected_rows = self.cursor.rowcount
                self.cursor.fetchone()
        except Exception:
            pass

        assert not affected_rows
        assert not result

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert db_span.ec == 2
        assert db_span.data["pg"]["error"] == (
            'relation "blah" does not exist\nLINE 1: SELECT * from blah\n                      ^\n'
        )

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert db_span.data["pg"]["stmt"] == "SELECT * from blah"
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    # Added to validate unicode support and register_type.
    def test_unicode(self) -> None:
        ext.register_type(ext.UNICODE, self.cursor)
        snowman = "\u2603"

        self.cursor.execute("delete from users where id in (1,2,3)")

        # unicode in statement
        psycopg2.extras.execute_batch(
            self.cursor,
            "insert into users (id, name) values (%%s, %%s) -- %s" % snowman,
            [(1, "x")],
        )
        self.cursor.execute("select id, name from users where id = 1")
        assert self.cursor.fetchone() == (1, "x")

        # unicode in data
        psycopg2.extras.execute_batch(
            self.cursor, "insert into users (id, name) values (%s, %s)", [(2, snowman)]
        )
        self.cursor.execute("select id, name from users where id = 2")
        assert self.cursor.fetchone() == (2, snowman)

        # unicode in both
        psycopg2.extras.execute_batch(
            self.cursor,
            "insert into users (id, name) values (%%s, %%s) -- %s" % snowman,
            [(3, snowman)],
        )
        self.cursor.execute("select id, name from users where id = 3")
        assert self.cursor.fetchone() == (3, snowman)

    def test_register_type(self) -> None:
        import uuid

        oid1 = 2950
        oid2 = 2951

        ext.UUID = ext.new_type(
            (oid1,), "UUID", lambda data, cursor: data and uuid.UUID(data) or None
        )
        ext.UUIDARRAY = ext.new_array_type((oid2,), "UUID[]", ext.UUID)

        ext.register_type(ext.UUID, self.cursor)
        ext.register_type(ext.UUIDARRAY, self.cursor)

    def test_connect_cursor_ctx_mgr(self) -> None:
        with tracer.start_as_current_span("test"):
            with self.db as connection:
                with connection.cursor() as cursor:
                    cursor.execute("""SELECT * from users""")
                    affected_rows = cursor.rowcount
                    result = cursor.fetchone()

        assert affected_rows == 1
        assert len(result) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert db_span.data["pg"]["stmt"] == "SELECT * from users"
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_connect_ctx_mgr(self) -> None:
        with tracer.start_as_current_span("test"):
            with self.db as connection:
                cursor = connection.cursor()
                cursor.execute("""SELECT * from users""")
                affected_rows = cursor.rowcount
                result = cursor.fetchone()

        assert affected_rows == 1
        assert len(result) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert db_span.data["pg"]["stmt"] == "SELECT * from users"
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_cursor_ctx_mgr(self) -> None:
        with tracer.start_as_current_span("test"):
            connection = self.db
            with connection.cursor() as cursor:
                cursor.execute("""SELECT * from users""")
                affected_rows = cursor.rowcount
                result = cursor.fetchone()

        assert affected_rows == 1
        assert len(result) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]
        assert db_span.data["pg"]["user"] == testenv["postgresql_user"]
        assert db_span.data["pg"]["stmt"] == "SELECT * from users"
        assert db_span.data["pg"]["host"] == testenv["postgresql_host"]
        assert db_span.data["pg"]["port"] == testenv["postgresql_port"]

    def test_deprecated_parameter_database(self) -> None:
        with tracer.start_as_current_span("test"):
            self.cursor.execute("""SELECT * from users""")
            affected_rows = self.cursor.rowcount
            result = self.cursor.fetchone()
            self.db.commit()

        assert affected_rows == 1
        assert len(result) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "postgres"
        assert db_span.data["pg"]["db"] == testenv["postgresql_db"]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018


import re
from typing import Any, Dict

from opentelemetry import context, trace

from instana.log import logger
from instana.span.span import InstanaSpan, get_current_span
from instana.span_context import SpanContext
from instana.util.traceutils import get_tracer_tuple, tracing_is_off

try:
    from sqlalchemy import __version__ as sqlalchemy_version
    from sqlalchemy import event
    from sqlalchemy.engine import Engine

    url_regexp = re.compile(r"\/\/(\S+@)")

    @event.listens_for(Engine, "before_cursor_execute", named=True)
    def receive_before_cursor_execute(
        **kw: Dict[str, Any],
    ) -> None:
        try:
            # If we're not tracing, just return
            if tracing_is_off():
                return

            tracer, parent_span, _ = get_tracer_tuple()
            parent_context = parent_span.get_span_context() if parent_span else None

            span = tracer.start_span("sqlalchemy", span_context=parent_context)
            conn = kw["conn"]
            conn.span = span
            span.set_attribute("sqlalchemy.sql", kw["statement"])
            span.set_attribute("sqlalchemy.eng", conn.engine.name)
            span.set_attribute(
                "sqlalchemy.url", url_regexp.sub("//", str(conn.engine.url))
            )

            ctx = trace.set_span_in_context(span)
            token = context.attach(ctx)
            conn.token = token
        except Exception:
            logger.debug(
                "Instrumenting sqlalchemy @ receive_before_cursor_execute",
                exc_info=True,
            )

    @event.listens_for(Engine, "after_cursor_execute", named=True)
    def receive_after_cursor_execute(
        **kw: Dict[str, Any],
    ) -> None:
        try:
            # If we're not tracing, just return
            if tracing_is_off():
                return

            current_span = get_current_span()
            conn = kw["conn"]
            if current_span.is_recording():
                current_span.end()
            if hasattr(conn, "token"):
                context.detach(conn.token)
                conn.token = None
        except Exception:
            logger.debug(
                "Instrumenting sqlalchemy @ receive_after_cursor_execute",
                exc_info=True,
            )

    error_event = "handle_error"
    # Handle dbapi_error event; deprecated since version 0.9
    if sqlalchemy_version[0] == "0":
        error_event = "dbapi_error"

    def _set_error_attributes(
        context: SpanContext,
        exception_string: str,
        span: InstanaSpan,
    ) -> None:
        context_exception = None, None
        if hasattr(context, exception_string):
            context_exception = getattr(context, exception_string)
        if span and context_exception:
            span.record_exception(context_exception)
        else:
            span.record_exception(f"No {error_event} specified.")
        if span.is_recording():
            span.end()

    @event.listens_for(Engine, error_event, named=True)
    def receive_handle_db_error(
        **kw: Dict[str, Any],
    ) -> None:
        try:
            if tracing_is_off():
                return

            current_span = get_current_span()

            # support older db error event
            if error_event == "dbapi_error":
                context = kw.get("context")
                exception_string = "exception"
            else:
                context = kw.get("exception_context")
                exception_string = "sqlalchemy_exception"

            if context:
                _set_error_attributes(context, exception_string, current_span)
        except Exception:
            logger.debug(
                "Instrumenting sqlalchemy @ receive_handle_db_error",
                exc_info=True,
            )

    logger.debug("Instrumenting sqlalchemy")

except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright IBM Corp. 2024

import logging

from wsgiref.simple_server import make_server
from bottle import default_app, response

from tests.helpers import testenv
from instana.middleware import InstanaWSGIMiddleware

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

testenv["wsgi_port"] = 10812
testenv["wsgi_server"] = ("http://127.0.0.1:" + str(testenv["wsgi_port"]))

app = default_app()

@app.route("/")
def hello():
    return "<center><h1>🐍 Hello Stan! 🦄</h1></center>"

@app.route("/response_headers")
def response_headers():
    response.set_header("X-Capture-This", "this")
    response.set_header("X-Capture-That", "that")
    return "Stan wuz here with headers!"

# Wrap the application with the Instana WSGI Middleware
app = InstanaWSGIMiddleware(app)
bottle_server = make_server('127.0.0.1', testenv["wsgi_port"], app)

if __name__ == "__main__":
    bottle_server.request_queue_size = 20
    bottle_server.serve_forever()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import typing

from opentelemetry.trace import SpanContext as OtelSpanContext
from opentelemetry.trace.span import (
    DEFAULT_TRACE_OPTIONS,
    DEFAULT_TRACE_STATE,
    TraceFlags,
    TraceState,
    format_span_id,
)


class SpanContext(OtelSpanContext):
    """The state of a Span to propagate between processes.

    This class includes the immutable attributes of a :class:`.Span` that must
    be propagated to a span's children and across process boundaries.

    Required Args:
        trace_id: The ID of the trace that this span belongs to.
        span_id: This span's ID.
        is_remote: True if propagated from a remote parent.
    """

    def __new__(
        cls,
        trace_id: int,
        span_id: int,
        is_remote: bool,
        trace_flags: typing.Optional[TraceFlags] = DEFAULT_TRACE_OPTIONS,
        trace_state: typing.Optional[TraceState] = DEFAULT_TRACE_STATE,
        level=1,
        synthetic=False,
        trace_parent=None,  # true/false flag,
        instana_ancestor=None,
        long_trace_id=None,
        correlation_type=None,
        correlation_id=None,
        traceparent=None,  # temporary storage of the validated traceparent header of the incoming request
        tracestate=None,  # temporary storage of the tracestate header
        **kwargs,
    ) -> "SpanContext":
        instance = super().__new__(cls, trace_id, span_id, is_remote, trace_flags, trace_state)
        return tuple.__new__(
            cls,
            (
                instance.trace_id,
                instance.span_id,
                instance.is_remote,
                instance.trace_flags,
                instance.trace_state,
                instance.is_valid,
                level,
                synthetic,
                trace_parent,  # true/false flag,
                instana_ancestor,
                long_trace_id,
                correlation_type,
                correlation_id,
                traceparent,  # temporary storage of the validated traceparent header of the incoming request
                tracestate,  # temporary storage of the tracestate header
            ),
        )

    def __getnewargs__(
        self,
    ):  # -> typing.Tuple[int, int, bool, "TraceFlags", "TraceState", int, bool, bool]:
        return (
            self.trace_id,
            self.span_id,
            self.is_remote,
            self.trace_flags,
            self.trace_state,
            self.level,
            self.synthetic,
            self.trace_parent,
            self.instana_ancestor,
            self.long_trace_id,
            self.correlation_type,
            self.correlation_id,
            self.traceparent,
            self.tracestate,
        )

    @property
    def level(self) -> int:
        return self[6]

    @property
    def synthetic(self) -> bool:
        return self[7]

    @property
    def trace_parent(self) -> bool:
        return self[8]

    @property
    def instana_ancestor(self):
        return self[9]

    @property
    def long_trace_id(self):
        return self[10]

    @property
    def correlation_type(self):
        return self[11]

    @property
    def correlation_id(self):
        return self[12]

    @property
    def traceparent(self):
        return self[13]

    @property
    def tracestate(self):
        return self[14]

    @property
    def suppression(self) -> bool:
        return self.level == 0

    def __repr__(self) -> str:
        return f"{type(self).__name__}(trace_id=0x{format_span_id(self.trace_id)}, span_id=0x{format_span_id(self.span_id)}, trace_flags=0x{self.trace_flags:02x}, trace_state={self.trace_state!r}, is_remote={self.is_remote}, synthetic={self.synthetic})"



# (c) Copyright IBM Corp. 2024

from typing import Generator
from unittest.mock import Mock

import pytest

from instana.propagators.base_propagator import BasePropagator


class TestBasePropagator:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.propagator = BasePropagator()
        yield
        self.propagator = None

    def test_extract_headers_dict(self) -> None:
        carrier_as_a_dict = {"key": "value"}
        assert carrier_as_a_dict == self.propagator.extract_headers_dict(
            carrier_as_a_dict
        )
        mocked_carrier = Mock()
        mocked_carrier.__dict__ = carrier_as_a_dict
        assert carrier_as_a_dict == self.propagator.extract_headers_dict(mocked_carrier)
        wrong_carrier = "value"
        assert self.propagator.extract_headers_dict(wrong_carrier) is None

    def test_get_ctx_level(self) -> None:
        assert 3 == self.propagator._get_ctx_level("3,4")
        assert 1 == self.propagator._get_ctx_level("wrong_data")

    def test_get_correlation_properties(self) -> None:
        a, b = self.propagator._get_correlation_properties(
            ",correlationType=3;correlationId=5;"
        )
        assert a == "3"
        assert b == "5"
        assert "3", None == self.propagator._get_correlation_properties(  # noqa: E711
            ",correlationType=3;"
        )

    def test_get_participating_trace_context(self, span_context) -> None:
        traceparent, tracestate = self.propagator._get_participating_trace_context(
            span_context
        )
        assert traceparent == "00-00000000000000001926b88ec9ee75ab-5fb1cff576b7e2f5-01"
        assert tracestate == "in=1926b88ec9ee75ab;5fb1cff576b7e2f5"

    def test_extract_instana_headers(self) -> None:
        dc = {
            "x-instana-t": "123456789",
            "x-instana-s": "12345",
            "x-instana-l": str.encode(",correlationType=3;correlationId=5;"),
            "x-instana-synthetic": "1",
        }
        trace_id, span_id, level, synthetic = self.propagator.extract_instana_headers(
            dc=dc
        )
        assert trace_id == "123456789"
        assert span_id == "12345"
        assert level == ",correlationType=3;correlationId=5;"
        assert synthetic

    def test_extract(self) -> None:
        carrier = {
            "x-instana-t": "123456789",
            "x-instana-s": "12345",
            "x-instana-l": str.encode("3,correlationId=5;"),
            "x-instana-synthetic": "1",
            "traceparent": "00-1812338823475918251-6895521157646639861-01",
            "tracestate": "in=1812338823475918251;6895521157646639861",
        }
        span_context = self.propagator.extract(
            carrier=carrier, disable_w3c_trace_context=True
        )
        assert span_context
        span_context = self.propagator.extract(carrier=carrier)
        span_context = self.propagator.extract(
            carrier=None, disable_w3c_trace_context=True
        )
        assert not span_context
        carrier.pop("x-instana-t", None)
        carrier.pop("x-instana-s", None)
        span_context = self.propagator.extract(carrier=carrier)
        assert span_context
        carrier = {
            "x-instana-t": "123456789",
            "x-instana-s": "12345",
            "x-instana-l": "2,correlationType=3;correlationId=5;",
            "x-instana-synthetic": "1",
            "traceparent": "00-4bf92f3577b34da61234567899999999-1234567890888888-01",
            "tracestate": "in=1812338823475918251;6895521157646639861",
        }
        span_context = self.propagator.extract(carrier=carrier)
        assert span_context



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Tuple
import wrapt

from opentelemetry.semconv.trace import SpanAttributes

from instana.log import logger
from instana.propagators.format import Format
from instana.singletons import agent
from instana.util.secrets import strip_secrets_from_query
from instana.util.traceutils import get_tracer_tuple, tracing_is_off, extract_custom_headers

try:
    import aiohttp

    if TYPE_CHECKING:
        from aiohttp.client import ClientSession
        from instana.span.span import InstanaSpan


    async def stan_request_start(
        session: "ClientSession", trace_config_ctx: SimpleNamespace, params
    ) -> Awaitable[None]:
        try:
            # If we're not tracing, just return
            if tracing_is_off():
                trace_config_ctx.span_context = None
                return

            tracer, parent_span, _ = get_tracer_tuple()
            parent_context = parent_span.get_span_context() if parent_span else None

            span = tracer.start_span("aiohttp-client", span_context=parent_context)

            extract_custom_headers(span, params.headers)

            tracer.inject(span.context, Format.HTTP_HEADERS, params.headers)

            parts = str(params.url).split("?")
            if len(parts) > 1:
                cleaned_qp = strip_secrets_from_query(
                    parts[1], agent.options.secrets_matcher, agent.options.secrets_list
                )
                span.set_attribute("http.params", cleaned_qp)
            span.set_attribute(SpanAttributes.HTTP_URL, parts[0])
            span.set_attribute(SpanAttributes.HTTP_METHOD, params.method)
            trace_config_ctx.span_context = span
        except Exception:
            logger.debug("aiohttp-client stan_request_start error:", exc_info=True)

    async def stan_request_end(
        session: "ClientSession", trace_config_ctx: SimpleNamespace, params
    ) -> Awaitable[None]:
        try:
            span: "InstanaSpan" = trace_config_ctx.span_context
            if span:
                span.set_attribute(
                    SpanAttributes.HTTP_STATUS_CODE, params.response.status
                )

                extract_custom_headers(span, params.response.headers)

                if 500 <= params.response.status:
                    span.mark_as_errored({"http.error": params.response.reason})

                if span.is_recording():
                    span.end()
                trace_config_ctx = None
        except Exception:
            logger.debug("aiohttp-client stan_request_end error:", exc_info=True)

    async def stan_request_exception(
        session: "ClientSession", trace_config_ctx: SimpleNamespace, params
    ) -> Awaitable[None]:
        try:
            span: "InstanaSpan" = trace_config_ctx.span_context
            if span:
                span.record_exception(params.exception)
                span.set_attribute("http.error", str(params.exception))
                if span.is_recording():
                    span.end()
                trace_config_ctx = None
        except Exception:
            logger.debug("aiohttp-client stan_request_exception error:", exc_info=True)

    @wrapt.patch_function_wrapper("aiohttp.client", "ClientSession.__init__")
    def init_with_instana(
        wrapped: Callable[..., Awaitable["ClientSession"]],
        instance: aiohttp.client.ClientSession,
        args: Tuple[int, str, Tuple[object, ...]],
        kwargs: Dict[str, Any],
    ) -> object:
        instana_trace_config = aiohttp.TraceConfig()
        instana_trace_config.on_request_start.append(stan_request_start)
        instana_trace_config.on_request_end.append(stan_request_end)
        instana_trace_config.on_request_exception.append(stan_request_exception)
        if "trace_configs" in kwargs:
            kwargs["trace_configs"].append(instana_trace_config)
        else:
            kwargs["trace_configs"] = [instana_trace_config]

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting aiohttp client")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
from tests.apps.pyramid.pyramid_app.app import pyramid_server as server
from tests.apps.utils import launch_background_thread

app_thread = None

if not os.environ.get("CASSANDRA_TEST"):
    app_thread = launch_background_thread(server.serve_forever, "Pyramid")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from typing import Generator

import pytest
from instana.singletons import agent, tracer
from starlette.testclient import TestClient

from instana.util.ids import hex_id
from tests.apps.starlette_app.app2 import starlette_server
from tests.helpers import get_first_span_by_filter


class TestStarletteMiddleware:
    """
    Tests Starlette with provided Middleware.
    """

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # We are using the TestClient from Starlette to make it easier.
        self.client = TestClient(starlette_server)
        # Clear all spans before a test run.
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        yield

    def test_vanilla_get(self) -> None:
        result = self.client.get("/")

        assert result
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        # Starlette instrumentation (like all instrumentation) _always_ traces
        # unless told otherwise
        spans = self.recorder.queued_spans()

        assert len(spans) == 1
        assert spans[0].n == "asgi"

    def test_basic_get(self) -> None:
        result = None
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/", headers=headers)

        assert result
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p

        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200
        assert asgi_span.data["http"]["host"] == "testserver"
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_basic_get_500(self) -> None:
        result = None
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/five", headers=headers)

        assert result
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p

        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert asgi_span.ec == 1
        assert asgi_span.data["http"]["path"] == "/five"
        assert asgi_span.data["http"]["path_tpl"] == "/five"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 500
        assert asgi_span.data["http"]["host"] == "testserver"
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from typing import Optional, Tuple, TYPE_CHECKING, Union, Dict, List, Any, Iterable

from instana.log import logger
from instana.singletons import agent, tracer
from instana.span.span import get_current_span
from instana.tracer import InstanaTracer

if TYPE_CHECKING:
    from instana.span.span import InstanaSpan

def extract_custom_headers(span: "InstanaSpan", headers: Optional[Union[Dict[str, Any], List[Tuple[object, ...]], Iterable]] = None, format: Optional[bool] = False) -> None:
    if not (agent.options.extra_http_headers and headers):
        return
    try:
        for custom_header in agent.options.extra_http_headers:
            # Headers are available in the following formats: HTTP_X_CAPTURE_THIS, b'x-header-1', X-Capture-That
            expected_header = (
                ("HTTP_" + custom_header.upper()).replace("-", "_")
                if format
                else custom_header
            )
            for header in headers:
                if isinstance(header, tuple):
                    header_key = header[0].decode("utf-8") if isinstance(header[0], bytes) else header[0]
                    header_val = header[1].decode("utf-8") if isinstance(header[1], bytes) else header[1]
                    if header_key.lower() == expected_header.lower():
                        span.set_attribute(
                            f"http.header.{custom_header}", header_val,
                        )  
                elif header.lower() == expected_header.lower():
                    span.set_attribute(f"http.header.{custom_header}", headers[expected_header])
    except Exception:
        logger.debug("extract_custom_headers: ", exc_info=True)


def get_active_tracer() -> Optional[InstanaTracer]:
    try:
        current_span = get_current_span()
        if current_span:
            # asyncio Spans are used as NonRecording Spans solely for context propagation
            if current_span.is_recording() or current_span.name == "asyncio":
                return tracer
            return None
        return None
    except Exception:
        # Do not try to log this with instana, as there is no active tracer and there will be an infinite loop at least
        # for PY2
        return None


def get_tracer_tuple() -> (
    Tuple[Optional[InstanaTracer], Optional["InstanaSpan"], Optional[str]]
):
    active_tracer = get_active_tracer()
    current_span = get_current_span()
    if active_tracer:
        return (active_tracer, current_span, current_span.name)
    elif agent.options.allow_exit_as_root:
        return (tracer, None, None)
    return (None, None, None)


def tracing_is_off() -> bool:
    return not (bool(get_active_tracer()) or agent.options.allow_exit_as_root)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import platform
import signal
import threading
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union

from instana.autoprofile.frame_cache import FrameCache
from instana.autoprofile.runtime import RuntimeInfo, min_version, register_signal
from instana.autoprofile.sampler_scheduler import SamplerConfig, SamplerScheduler
from instana.autoprofile.samplers.allocation_sampler import AllocationSampler
from instana.autoprofile.samplers.block_sampler import BlockSampler
from instana.autoprofile.samplers.cpu_sampler import CPUSampler
from instana.log import logger

if TYPE_CHECKING:
    from types import FrameType
    from instana.agent.host import HostAgent


class Profiler(object):
    def __init__(self, agent: "HostAgent") -> None:
        self.agent = agent
        self.profiler_started = False
        self.profiler_destroyed = False
        self.sampler_active = False
        self.main_thread_func = None
        self.frame_cache = FrameCache(self)
        self.options = None
        self.cpu_sampler_scheduler = self._create_sampler_scheduler(
            CPUSampler(self), "CPU sampler", 20, 5, 30, 20, 120
        )
        self.allocation_sampler_scheduler = self._create_sampler_scheduler(
            AllocationSampler(self), "Allocation sampler", 20, 5, 30, 20, 120
        )
        self.block_sampler_scheduler = self._create_sampler_scheduler(
            BlockSampler(self), "Block sampler", 20, 5, 30, 20, 120
        )

    def get_option(
        self, name: str, default_val: Optional[object] = None
    ) -> Optional[object]:
        if name not in self.options:
            return default_val
        else:
            return self.options[name]

    def start(self, **kwargs: Dict[str, Any]) -> None:
        if self.profiler_started:
            return

        try:
            if not min_version(3, 8):
                raise Exception("Supported Python versions 3.8 or higher.")

            if platform.python_implementation() != "CPython":
                raise Exception("Supported Python interpreter is CPython.")

            if self.profiler_destroyed:
                logger.warning("Destroyed profiler cannot be started.")
                return

            self.options = kwargs
            self.frame_cache.start()
            self.cpu_sampler_scheduler.setup()
            self.allocation_sampler_scheduler.setup()
            self.block_sampler_scheduler.setup()

            # execute main_thread_func in main thread on signal
            def _signal_handler(signum: signal.Signals, frame: "FrameType") -> bool:
                if self.main_thread_func:
                    func = self.main_thread_func
                    self.main_thread_func = None
                    try:
                        func()
                    except Exception:
                        logger.error("Error in signal handler function", exc_info=True)

                    return True

            if not RuntimeInfo.OS_WIN:
                register_signal(signal.SIGUSR2, _signal_handler)

            self.cpu_sampler_scheduler.start()
            self.allocation_sampler_scheduler.start()
            self.block_sampler_scheduler.start()

            self.profiler_started = True
            logger.debug("Profiler started.")
        except Exception:
            logger.error("Error starting profiler", exc_info=True)

    def destroy(self) -> None:
        if not self.profiler_started:
            logger.warning("Profiler has not been started.")
            return

        if self.profiler_destroyed:
            return

        self.frame_cache.stop()
        self.cpu_sampler_scheduler.stop()
        self.allocation_sampler_scheduler.stop()
        self.block_sampler_scheduler.stop()

        self.cpu_sampler_scheduler.destroy()
        self.allocation_sampler_scheduler.destroy()
        self.block_sampler_scheduler.destroy()

        self.profiler_destroyed = True
        logger.debug("Profiler destroyed.")

    def run_in_thread(self, func: Callable[..., object]) -> threading.Thread:
        def func_wrapper() -> None:
            try:
                func()
            except Exception:
                logger.error("Error in thread function", exc_info=True)

        t = threading.Thread(target=func_wrapper)
        t.start()
        return t

    def run_in_main_thread(self, func: Callable[..., object]) -> bool:
        if self.main_thread_func:
            return False

        self.main_thread_func = func
        os.kill(os.getpid(), signal.SIGUSR2)

        return True

    def _create_sampler_scheduler(
        self,
        sampler: Union["AllocationSampler", "BlockSampler", "CPUSampler"],
        log_prefix: str,
        max_profile_duration: int,
        max_span_duration: int,
        max_span_count: int,
        span_interval: int,
        report_interval: int,
    ) -> SamplerScheduler:
        config = SamplerConfig()
        config.log_prefix = log_prefix
        config.max_profile_duration = max_profile_duration
        config.max_span_duration = max_span_duration
        config.max_span_count = max_span_count
        config.span_interval = span_interval
        config.report_interval = report_interval

        return SamplerScheduler(self, sampler, config)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import signal
import sys
from typing import TYPE_CHECKING, Callable, Optional

if TYPE_CHECKING:
    from types import FrameType

class RuntimeInfo(object):
    OS_LINUX = sys.platform.startswith("linux")
    OS_DARWIN = sys.platform == "darwin"
    OS_WIN = sys.platform == "win32"
    GEVENT = False


try:
    import gevent

    if hasattr(gevent, "_threading"):
        RuntimeInfo.GEVENT = True
except ImportError:
    pass


def min_version(major: int, minor: Optional[int] = 0) -> bool:
    return sys.version_info.major == major and sys.version_info.minor >= minor


def register_signal(
    signal_number: signal.Signals,
    handler_func: Callable[..., object],
    once: Optional[bool] = False,
) -> None:
    prev_handler = None

    def _handler(signum: signal.Signals, frame: "FrameType") -> None:
        skip_prev = handler_func(signum, frame)

        if not skip_prev:
            if callable(prev_handler):
                if once:
                    signal.signal(signum, prev_handler)
                prev_handler(signum, frame)
            elif prev_handler == signal.SIG_DFL and once:
                signal.signal(signum, signal.SIG_DFL)
                os.kill(os.getpid(), signum)

    prev_handler = signal.signal(signal_number, _handler)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

import os
import sys
import time
import threading

if not any((os.environ.get('GEVENT_STARLETTE_TEST'),
            os.environ.get('CASSANDRA_TEST'),
            sys.version_info < (3, 5, 3))):
    # Background RPC application
    #
    # Spawn the background RPC app that the tests will throw
    # requests at.
    import tests.apps.grpc_server
    from .stan_server import StanServicer
    stan_servicer = StanServicer()
    rpc_server_thread = threading.Thread(target=stan_servicer.start_server)
    rpc_server_thread.daemon = True
    rpc_server_thread.name = "Background RPC app"
    print("Starting background RPC app...")
    rpc_server_thread.start()
    time.sleep(1)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

import unittest

from instana.util.secrets import strip_secrets_from_query


class TestSecrets(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_equals_ignore_case(self):
        matcher = 'equals-ignore-case'
        kwlist = ['two']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=<redacted>&THREE=&4='+'&five='okyeah'")

    def test_equals(self):
        matcher = 'equals'
        kwlist = ['Two']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=<redacted>&THREE=&4='+'&five='okyeah'")

    def test_equals_no_match(self):
        matcher = 'equals'
        kwlist = ['two']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five='okyeah'")

    def test_contains_ignore_case(self):
        matcher = 'contains-ignore-case'
        kwlist = ['FI']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five=<redacted>")

    def test_contains_ignore_case_no_match(self):
        matcher = 'contains-ignore-case'
        kwlist = ['XXXXXX']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five='okyeah'")

    def test_contains(self):
        matcher = 'contains'
        kwlist = ['fi']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five=<redacted>")

    def test_contains_no_match(self):
        matcher = 'contains'
        kwlist = ['XXXXXX']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five='okyeah'")

    def test_regex(self):
        matcher = 'regex'
        kwlist = [r"\d"]

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4=<redacted>&five='okyeah'")

    def test_regex_no_match(self):
        matcher = 'regex'
        kwlist = [r"\d\d\d"]

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five='okyeah'")

    def test_equals_with_path_component(self):
        matcher = 'equals'
        kwlist = ['Two']

        query_params = "/signup?one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "/signup?one=1&Two=<redacted>&THREE=&4='+'&five='okyeah'")

    def test_equals_with_full_url(self):
        matcher = 'equals'
        kwlist = ['Two']

        query_params = "http://www.x.org/signup?one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "http://www.x.org/signup?one=1&Two=<redacted>&THREE=&4='+'&five='okyeah'")

    def test_equals_with_none(self):
        matcher = 'equals'
        kwlist = ['Two']

        query_params = None

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual('', stripped)

    def test_bad_matcher(self):
        matcher = 'BADCAFE'
        kwlist = ['Two']

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five='okyeah'")

    def test_bad_kwlist(self):
        matcher = 'equals'
        kwlist = None

        query_params = "one=1&Two=two&THREE=&4='+'&five='okyeah'"

        stripped = strip_secrets_from_query(query_params, matcher, kwlist)

        self.assertEqual(stripped, "one=1&Two=two&THREE=&4='+'&five='okyeah'")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: stan.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 27, 2, "", "stan.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nstan.proto\x12\x04stan"#\n\x0fQuestionRequest\x12\x10\n\x08question\x18\x01 \x01(\t"8\n\x10QuestionResponse\x12\x0e\n\x06\x61nswer\x18\x01 \x01(\t\x12\x14\n\x0cwas_answered\x18\x02 \x01(\x08\x32\xe3\x03\n\x04Stan\x12I\n\x16OneQuestionOneResponse\x12\x15.stan.QuestionRequest\x1a\x16.stan.QuestionResponse"\x00\x12M\n\x18ManyQuestionsOneResponse\x12\x15.stan.QuestionRequest\x1a\x16.stan.QuestionResponse"\x00(\x01\x12M\n\x18OneQuestionManyResponses\x12\x15.stan.QuestionRequest\x1a\x16.stan.QuestionResponse"\x00\x30\x01\x12P\n\x19ManyQuestionsManyReponses\x12\x15.stan.QuestionRequest\x1a\x16.stan.QuestionResponse"\x00(\x01\x30\x01\x12N\n\x1bOneQuestionOneErrorResponse\x12\x15.stan.QuestionRequest\x1a\x16.stan.QuestionResponse"\x00\x12P\n\x1dOneErroredQuestionOneResponse\x12\x15.stan.QuestionRequest\x1a\x16.stan.QuestionResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "stan_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_QUESTIONREQUEST"]._serialized_start = 20
    _globals["_QUESTIONREQUEST"]._serialized_end = 55
    _globals["_QUESTIONRESPONSE"]._serialized_start = 57
    _globals["_QUESTIONRESPONSE"]._serialized_end = 113
    _globals["_STAN"]._serialized_start = 116
    _globals["_STAN"]._serialized_end = 599
# @@protoc_insertion_point(module_scope)



# (c) Copyright IBM Corp. 2024

from typing import Generator
from unittest.mock import Mock, patch

import pytest

from instana.recorder import StanRecorder
from instana.span.base_span import BaseSpan
from instana.span.span import InstanaSpan
from instana.span_context import SpanContext
from instana.util import DictionaryOfStan


def test_basespan(
    span: InstanaSpan,
    trace_id: int,
    span_id: int,
) -> None:
    base_span = BaseSpan(span, None)

    expected_dict = {
        "t": trace_id,
        "p": None,
        "s": span_id,
        "ts": round(span.start_time / 10**6),
        "d": None,
        "f": None,
        "ec": None,
        "data": DictionaryOfStan(),
        "stack": None,
    }

    assert expected_dict["t"] == base_span.t
    assert expected_dict["s"] == base_span.s
    assert expected_dict["p"] == base_span.p
    assert expected_dict["ts"] == base_span.ts
    assert expected_dict["d"] == base_span.d
    assert not base_span.f
    assert expected_dict["ec"] == base_span.ec
    assert isinstance(base_span.data, dict)
    assert expected_dict["stack"] == base_span.stack
    assert not base_span.sy

    expected_dict_str = str(expected_dict)
    assert expected_dict_str == repr(base_span)
    assert f"BaseSpan({expected_dict_str})" == str(base_span)


def test_basespan_with_synthetic_source_and_kwargs(
    span: InstanaSpan,
    trace_id: int,
    span_id: int,
) -> None:
    span.synthetic = True
    source = "source test"
    _kwarg1 = "value1"
    base_span = BaseSpan(span, source, arg1=_kwarg1)

    assert trace_id == base_span.t
    assert span_id == base_span.s
    # synthetic should be true only for entry spans
    assert not base_span.sy
    assert source == base_span.f
    assert _kwarg1 == base_span.arg1


def test_populate_extra_span_attributes(
    span: InstanaSpan,
) -> None:
    base_span = BaseSpan(span, None)
    base_span._populate_extra_span_attributes(span)

    assert not hasattr(base_span, "tp")
    assert not hasattr(base_span, "tp")
    assert not hasattr(base_span, "ia")
    assert not hasattr(base_span, "lt")
    assert not hasattr(base_span, "crtp")
    assert not hasattr(base_span, "crid")


def test_populate_extra_span_attributes_with_values(
    trace_id: int,
    span_id: int,
    span_processor: StanRecorder,
) -> None:
    long_id = 1512366075204170929049582354406559215
    span_context = SpanContext(
        trace_id=trace_id,
        span_id=span_id,
        is_remote=False,
        synthetic=True,
        trace_parent=True,
        instana_ancestor="IDK",
        long_trace_id=long_id,
        correlation_type="IDK",
        correlation_id=long_id,
    )
    span = InstanaSpan("test-base-span", span_context, span_processor)
    base_span = BaseSpan(span, None)
    base_span._populate_extra_span_attributes(span)

    assert trace_id == base_span.t
    assert span_id == base_span.s
    # synthetic should be true only for entry spans
    assert not base_span.sy
    assert base_span.tp
    assert "IDK" == base_span.ia
    assert long_id == base_span.lt
    assert "IDK" == base_span.crtp
    assert long_id == base_span.crid


def test_validate_attributes(
    base_span: BaseSpan,
) -> None:
    attributes = {
        "field1": 1,
        "field2": "two",
    }
    filtered_attributes = base_span._validate_attributes(attributes)

    assert isinstance(filtered_attributes, dict)
    assert len(attributes) == len(filtered_attributes)
    for key, value in attributes.items():
        assert key in filtered_attributes.keys()
        assert value in filtered_attributes.values()


def test_validate_attribute_with_invalid_key_type(
    base_span: BaseSpan,
) -> None:
    key = 1
    value = "one"

    (validated_key, validated_value) = base_span._validate_attribute(key, value)

    assert not validated_key
    assert not validated_value


def test_validate_attribute_exception(
    span: InstanaSpan,
) -> None:
    base_span = BaseSpan(span, None)
    key = "field1"
    value = span

    with patch(
        "instana.span.base_span.BaseSpan._convert_attribute_value",
        side_effect=Exception("mocked error"),
    ):
        (validated_key, validated_value) = base_span._validate_attribute(key, value)
        assert key == validated_key
        assert not validated_value


def test_convert_attribute_value(
    span: InstanaSpan,
) -> None:
    base_span = BaseSpan(span, None)
    value = span

    converted_value = base_span._convert_attribute_value(value)
    assert "<instana.span.span.InstanaSpan object" in converted_value


def test_convert_attribute_value_exception(
    base_span: BaseSpan,
) -> None:
    mock = Mock()
    mock.__repr__ = Mock(side_effect=Exception("mocked error"))

    converted_value = base_span._convert_attribute_value(mock)
    assert not converted_value



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

try:
    from instana.log import logger
    from instana.singletons import agent

    import gunicorn
    from gunicorn.arbiter import Arbiter
    from gunicorn.config import Config
    from gunicorn.workers.sync import SyncWorker

    def pre_fork(config: Config, server: Arbiter, worker: SyncWorker) -> None:
        """This is our gunicorn hook to detect and act when worker processes are forked off."""
        logger.debug("Handling gunicorn fork...")
        agent.handle_fork()

    Config.pre_fork = pre_fork

    logger.debug("Gunicorn pre-fork hook applied")
except ImportError:
    logger.debug(
        "gunicorn hooks: decorators not available: likely not running under gunicorn"
    )
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021


from sanic.views import HTTPMethodView
from sanic.response import text


class SimpleView(HTTPMethodView):
    def get(self, request):
        return text("I am get method")

    # You can also use async syntax
    async def post(self, request):
        return text("I am post method")

    def put(self, request):
        return text("I am put method")

    def patch(self, request):
        return text("I am patch method")

    def delete(self, request):
        return text("I am delete method")



# (c) Copyright IBM Corp. 2024

import pytest

from instana.singletons import agent, tracer
from instana.tracer import InstanaTracer
from instana.util.traceutils import (
    extract_custom_headers,
    get_active_tracer,
    get_tracer_tuple,
    tracing_is_off,
)


@pytest.mark.parametrize(
    "custom_headers, format",
    [
        (
            {
                "X-Capture-This-Too": "this too",
                "X-Capture-That-Too": "that too",
            },
            False,
        ),
        (
            {
                "HTTP_X_CAPTURE_THIS_TOO": "this too",
                "HTTP_X_CAPTURE_THAT_TOO": "that too",
            },
            True,
        ),
        (
            [("X-CAPTURE-THIS-TOO", "this too"), ("x-capture-that-too", "that too")],
            False,
        ),
        (
            [
                (b"X-Capture-This-Too", b"this too"),
                (b"X-Capture-That-Too", b"that too"),
            ],
            False,
        ),
        (
            [
                ("HTTP_X_CAPTURE_THIS_TOO", "this too"),
                ("HTTP_X_CAPTURE_THAT_TOO", "that too"),
            ],
            True,
        ),
    ],
)
def test_extract_custom_headers(span, custom_headers, format) -> None:
    agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]
    extract_custom_headers(span, custom_headers, format=format)
    assert len(span.attributes) == 2
    assert span.attributes["http.header.X-Capture-This-Too"] == "this too"
    assert span.attributes["http.header.X-Capture-That-Too"] == "that too"


def test_get_activate_tracer(mocker) -> None:
    assert not get_active_tracer()

    with tracer.start_as_current_span("test"):
        response = get_active_tracer()
        assert isinstance(response, InstanaTracer)
        assert response == tracer
        with mocker.patch(
            "instana.span.span.InstanaSpan.is_recording", return_value=False
        ):
            assert not get_active_tracer()


def test_get_tracer_tuple() -> None:
    response = get_tracer_tuple()
    assert response == (None, None, None)

    agent.options.allow_exit_as_root = True
    response = get_tracer_tuple()
    assert response == (tracer, None, None)
    agent.options.allow_exit_as_root = False

    with tracer.start_as_current_span("test") as span:
        response = get_tracer_tuple()
        assert response == (tracer, span, span.name)


def test_tracing_is_off() -> None:
    response = tracing_is_off()
    assert response
    with tracer.start_as_current_span("test"):
        response = tracing_is_off()
        assert not response

    agent.options.allow_exit_as_root = True
    response = tracing_is_off()
    assert not response
    agent.options.allow_exit_as_root = False



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2016

# Accept, process and queue spans for eventual reporting.

import os
import queue
from typing import TYPE_CHECKING, List, Optional, Type

from instana.span.kind import REGISTERED_SPANS
from instana.span.readable_span import ReadableSpan
from instana.span.registered_span import RegisteredSpan
from instana.span.sdk_span import SDKSpan

if TYPE_CHECKING:
    from instana.agent.base import BaseAgent


class StanRecorder(object):
    THREAD_NAME = "InstanaSpan Recorder"

    # Recorder thread for collection/reporting of spans
    thread = None

    def __init__(self, agent: Optional[Type["BaseAgent"]] = None) -> None:
        if agent is None:
            # Late import to avoid circular import
            # pylint: disable=import-outside-toplevel
            from instana.singletons import get_agent

            self.agent = get_agent()
        else:
            self.agent = agent

    def queue_size(self) -> int:
        """Return the size of the queue; how may spans are queued,"""
        return self.agent.collector.span_queue.qsize()

    def queued_spans(self) -> List[ReadableSpan]:
        """Get all of the spans in the queue."""
        span = None
        spans = []

        if self.agent.collector.span_queue.empty() is True:
            return spans

        while True:
            try:
                span = self.agent.collector.span_queue.get(False)
            except queue.Empty:
                break
            else:
                spans.append(span)
        return spans

    def clear_spans(self):
        """Clear the queue of spans."""
        if not self.agent.collector.span_queue.empty():
            self.queued_spans()

    def record_span(self, span: ReadableSpan) -> None:
        """
        Convert the passed span into JSON and add it to the span queue.
        """
        if span.context.suppression:
            return

        if self.agent.can_send():
            service_name = None
            source = self.agent.get_from_structure()
            if "INSTANA_SERVICE_NAME" in os.environ:
                service_name = self.agent.options.service_name

            if span.name in REGISTERED_SPANS:
                json_span = RegisteredSpan(span, source, service_name)
            else:
                service_name = self.agent.options.service_name
                json_span = SDKSpan(span, source, service_name)

            # logger.debug("Recorded span: %s", json_span)
            self.agent.collector.span_queue.put(json_span)



# (c) Copyright IBM Corp. 2024

import logging
import queue
import threading
import time
from typing import Generator
from unittest.mock import patch

import pytest
from pytest import LogCaptureFixture

from instana.agent.host import HostAgent
from instana.collector.base import BaseCollector
from instana.recorder import StanRecorder
from instana.span.registered_span import RegisteredSpan
from instana.span.span import InstanaSpan
from instana.span_context import SpanContext


class TestBaseCollector:
    @pytest.fixture(autouse=True)
    def _resource(
        self,
        caplog: LogCaptureFixture,
    ) -> Generator[None, None, None]:
        self.collector = BaseCollector(HostAgent())
        yield
        self.collector.shutdown(report_final=False)
        self.collector = None
        caplog.clear()

    def test_default(self) -> None:
        assert isinstance(self.collector.agent, HostAgent)
        assert self.collector.THREAD_NAME == "Instana Collector"
        assert isinstance(self.collector.span_queue, queue.Queue)
        assert isinstance(self.collector.profile_queue, queue.Queue)
        assert not self.collector.reporting_thread
        assert isinstance(self.collector.thread_shutdown, threading.Event)
        assert self.collector.snapshot_data_last_sent == 0
        assert self.collector.snapshot_data_interval == 300
        assert len(self.collector.helpers) == 0
        assert self.collector.report_interval == 1
        assert not self.collector.started
        assert self.collector.fetching_start_time == 0

    def test_is_reporting_thread_running(self) -> None:
        stop_event = threading.Event()

        def reporting_function():
            stop_event.wait()

        sample_thread = threading.Thread(
            name=self.collector.THREAD_NAME, target=reporting_function
        )
        sample_thread.start()
        try:
            assert self.collector.is_reporting_thread_running()
        finally:
            stop_event.set()
            sample_thread.join()

    def test_is_reporting_thread_running_with_different_name(self) -> None:
        self.collector.THREAD_NAME = "sample-collector"
        stop_event = threading.Event()

        def reporting_function():
            stop_event.wait()

        sample_thread = threading.Thread(name="test-thread", target=reporting_function)
        sample_thread.start()
        try:
            assert not self.collector.is_reporting_thread_running()
        finally:
            stop_event.set()
            sample_thread.join()

    def test_start_collector_while_running_thread(
        self,
        caplog: LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        with patch(
            "instana.collector.base.BaseCollector.is_reporting_thread_running",
            return_value=True,
        ):
            self.collector.start()
            assert (
                "BaseCollector.start non-fatal: call but thread already running (started: False)"
                in caplog.messages
            )

    def test_start_agent_shutdown_is_set(self) -> None:
        self.collector.thread_shutdown.set()
        isThreadFound = False
        with patch(
            "instana.collector.base.BaseCollector.is_reporting_thread_running",
            return_value=True,
        ):
            response = self.collector.start()
            assert not response
            for thread in threading.enumerate():
                if thread.name == "Collector Timed Start":
                    isThreadFound = True
            assert isThreadFound

    def test_start_collector_when_agent_is_ready(
        self,
        caplog: LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        with patch(
            "instana.collector.base.BaseCollector.is_reporting_thread_running",
            return_value=False,
        ):
            if not self.collector.started:
                self.collector.start()
                assert self.collector.started
                assert self.collector.reporting_thread.daemon
                assert (
                    self.collector.reporting_thread.name == self.collector.THREAD_NAME
                )

    def test_start_agent_can_not_send(
        self,
        caplog: LogCaptureFixture,
    ) -> None:
        with patch(
            "instana.collector.base.BaseCollector.is_reporting_thread_running",
            return_value=False,
        ), patch("instana.agent.host.HostAgent.can_send", return_value=False):
            caplog.set_level(logging.WARNING, logger="instana")
            self.collector.agent.machine.fsm.current = "test"
            self.collector.start()
            assert (
                "BaseCollector.start: the agent tells us we can't send anything out"
                in caplog.messages
            )

    def test_shutdown(
        self,
        caplog: LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        self.collector.shutdown()
        assert "Collector.shutdown: Reporting final data." in caplog.messages
        assert not self.collector.started

    def test_background_report(self) -> None:
        assert self.collector.background_report()
        self.collector.thread_shutdown.set()
        assert not self.collector.background_report()

    def test_should_send_snapshot_data(self, caplog: LogCaptureFixture) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        self.collector.should_send_snapshot_data()
        assert (
            "BaseCollector: should_send_snapshot_data needs to be overridden"
            in caplog.messages
        )

    def test_collect_snapshot(
        self,
        caplog: LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        self.collector.collect_snapshot()
        assert (
            "BaseCollector: collect_snapshot needs to be overridden" in caplog.messages
        )

    def test_queued_spans(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_list = [
            RegisteredSpan(
                InstanaSpan("span1", span_context, span_processor), None, "log"
            ),
            RegisteredSpan(
                InstanaSpan("span2", span_context, span_processor), None, "log"
            ),
            RegisteredSpan(
                InstanaSpan("span3", span_context, span_processor), None, "log"
            ),
        ]
        for span in span_list:
            self.collector.span_queue.put(span)
        time.sleep(0.1)
        spans = self.collector.queued_spans()
        assert len(spans) == 3

    def test_queued_profiles(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_list = [
            RegisteredSpan(
                InstanaSpan("span1", span_context, span_processor), None, "log"
            ),
            RegisteredSpan(
                InstanaSpan("span2", span_context, span_processor), None, "log"
            ),
            RegisteredSpan(
                InstanaSpan("span3", span_context, span_processor), None, "log"
            ),
        ]
        for span in span_list:
            self.collector.profile_queue.put(span)
        time.sleep(0.1)
        profiles = self.collector.queued_profiles()
        assert len(profiles) == 3



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

"""
couchbase instrumentation - This instrumentation supports the Python CouchBase 2.3.4 --> 2.5.x SDK currently:
https://docs.couchbase.com/python-sdk/2.5/start-using-sdk.html
"""

try:
    import couchbase
    from instana.log import logger

    if not (
        hasattr(couchbase, "__version__")
        and (couchbase.__version__ >= "2.3.4" and couchbase.__version__ < "3.0.0")
    ):
        logger.debug("Instana supports 2.3.4 <= couchbase_versions < 3.0.0. Skipping.")
        raise ImportError

    from couchbase.bucket import Bucket
    from couchbase.n1ql import N1QLQuery

    from typing import Any, Callable, Dict, Tuple, Union

    import wrapt

    from instana.span.span import InstanaSpan
    from instana.util.traceutils import get_tracer_tuple, tracing_is_off

    # List of operations to instrument
    # incr, incr_multi, decr, decr_multi, retrieve_in are wrappers around operations above
    operations = [
        "upsert",
        "insert",
        "replace",
        "append",
        "prepend",
        "get",
        "rget",
        "touch",
        "lock",
        "unlock",
        "remove",
        "counter",
        "mutate_in",
        "lookup_in",
        "stats",
        "ping",
        "diagnostics",
        "observe",
        "upsert_multi",
        "insert_multi",
        "replace_multi",
        "append_multi",
        "prepend_multi",
        "get_multi",
        "touch_multi",
        "lock_multi",
        "unlock_multi",
        "observe_multi",
        "endure_multi",
        "remove_multi",
        "counter_multi",
    ]

    def collect_attributes(
        span: InstanaSpan,
        instance: Bucket,
        query_arg: Union[N1QLQuery, object],
        op: str,
    ) -> None:
        try:
            span.set_attribute("couchbase.hostname", instance.server_nodes[0])
            span.set_attribute("couchbase.bucket", instance.bucket)
            span.set_attribute("couchbase.type", op)

            if query_arg:
                query = None
                if type(query_arg) is N1QLQuery:
                    query = query_arg.statement
                else:
                    query = query_arg

                span.set_attribute("couchbase.sql", query)
        except Exception:
            # No fail on key capture - best effort
            pass

    def make_wrapper(op: str) -> Callable:
        def wrapper(
            wrapped: Callable[..., object],
            instance: couchbase.bucket.Bucket,
            args: Tuple[object, ...],
            kwargs: Dict[str, Any],
        ) -> object:
            tracer, parent_span, _ = get_tracer_tuple()
            parent_context = parent_span.get_span_context() if parent_span else None

            # If we're not tracing, just return
            if tracing_is_off():
                return wrapped(*args, **kwargs)

            with tracer.start_as_current_span(
                "couchbase", span_context=parent_context
            ) as span:
                collect_attributes(span, instance, None, op)
                try:
                    return wrapped(*args, **kwargs)
                except Exception as exc:
                    span.record_exception(exc)
                    span.set_attribute("couchbase.error", repr(exc))
                    logger.debug("Instana couchbase @ wrapper", exc_info=True)

        return wrapper

    def query_with_instana(
        wrapped: Callable[..., object],
        instance: couchbase.bucket.Bucket,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        # If we're not tracing, just return
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        with tracer.start_as_current_span(
            "couchbase", span_context=parent_context
        ) as span:
            try:
                collect_attributes(span, instance, args[0], "n1ql_query")
                return wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
                span.set_attribute("couchbase.error", repr(exc))
                logger.debug("Instana couchbase @ query_with_instana", exc_info=True)

    logger.debug("Instrumenting couchbase")
    wrapt.wrap_function_wrapper(
        "couchbase.bucket", "Bucket.n1ql_query", query_with_instana
    )
    for op in operations:
        f = make_wrapper(op)
        wrapt.wrap_function_wrapper("couchbase.bucket", f"Bucket.{op}", f)

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import instana

from sanic import Sanic
from sanic.exceptions import SanicException
from sanic.response import text

from tests.apps.sanic_app.simpleview import SimpleView
from tests.apps.sanic_app.name import NameView

app = Sanic("test")


@app.get("/foo/<foo_id:int>")
async def uuid_handler(request, foo_id: int):
    return text("INT - {}".format(foo_id))


@app.route("/response_headers")
async def response_headers(request):
    headers = {"X-Capture-This-Too": "this too", "X-Capture-That-Too": "that too"}
    return text("Stan wuz here with headers!", headers=headers)


@app.route("/test_request_args")
async def test_request_args_500(request):
    raise SanicException("Something went wrong.", status_code=500)


@app.route("/instana_exception")
async def test_instana_exception(request):
    raise SanicException(description="Something went wrong.", status_code=500)


@app.route("/wrong")
async def test_request_args_400(request):
    raise SanicException(message="Something went wrong.", status_code=400)


@app.get("/tag/<tag>")
async def tag_handler(request, tag):
    return text("Tag - {}".format(tag))


app.add_route(SimpleView.as_view(), "/")
app.add_route(NameView.as_view(), "/<name>")


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=8000, debug=True, access_log=True)



# (c) Copyright IBM Corp. 2024

from instana.log import logger
from instana.span.base_span import BaseSpan
from instana.span.kind import ENTRY_SPANS, EXIT_SPANS, HTTP_SPANS, LOCAL_SPANS

from opentelemetry.trace import SpanKind
from opentelemetry.semconv.trace import SpanAttributes


class RegisteredSpan(BaseSpan):
    def __init__(self, span, source, service_name, **kwargs) -> None:
        # pylint: disable=invalid-name
        super(RegisteredSpan, self).__init__(span, source, **kwargs)
        self.n = span.name
        self.k = SpanKind.SERVER  # entry -> Server span represents a synchronous incoming remote call such as an incoming HTTP request

        self.data["service"] = service_name
        if span.name in ENTRY_SPANS:
            # entry
            self._populate_entry_span_data(span)
            self._populate_extra_span_attributes(span)
        elif span.name in EXIT_SPANS:
            self.k = SpanKind.CLIENT  # exit -> Client span represents a synchronous outgoing remote call such as an outgoing HTTP request or database call
            self._populate_exit_span_data(span)
        elif span.name in LOCAL_SPANS:
            self.k = SpanKind.INTERNAL  # intermediate -> Internal span represents an internal operation within an application
            self._populate_local_span_data(span)

        if "rabbitmq" in self.data and self.data["rabbitmq"]["sort"] == "publish":
            self.k = SpanKind.CLIENT  # exit

        # unify the span name for gcps-producer and gcps-consumer
        if "gcps" in span.name:
            self.n = "gcps"

        # Store any leftover attributes in the custom section
        if len(span.attributes) > 0:
            self.data["custom"]["attributes"] = self._validate_attributes(
                span.attributes
            )

    def _populate_entry_span_data(self, span) -> None:
        if span.name in HTTP_SPANS:
            self._collect_http_attributes(span)

        elif span.name == "aws.lambda.entry":
            self.data["lambda"]["arn"] = span.attributes.pop("lambda.arn", "Unknown")
            self.data["lambda"]["alias"] = None
            self.data["lambda"]["runtime"] = "python"
            self.data["lambda"]["functionName"] = span.attributes.pop(
                "lambda.name", "Unknown"
            )
            self.data["lambda"]["functionVersion"] = span.attributes.pop(
                "lambda.version", "Unknown"
            )
            self.data["lambda"]["trigger"] = span.attributes.pop("lambda.trigger", None)
            self.data["lambda"]["error"] = span.attributes.pop("lambda.error", None)

            trigger_type = self.data["lambda"]["trigger"]

            if trigger_type in ["aws:api.gateway", "aws:application.load.balancer"]:
                self._collect_http_attributes(span)
            elif trigger_type == "aws:cloudwatch.events":
                self.data["lambda"]["cw"]["events"]["id"] = span.attributes.pop(
                    "data.lambda.cw.events.id", None
                )
                self.data["lambda"]["cw"]["events"]["more"] = span.attributes.pop(
                    "lambda.cw.events.more", False
                )
                self.data["lambda"]["cw"]["events"]["resources"] = span.attributes.pop(
                    "lambda.cw.events.resources", None
                )

            elif trigger_type == "aws:cloudwatch.logs":
                self.data["lambda"]["cw"]["logs"]["group"] = span.attributes.pop(
                    "lambda.cw.logs.group", None
                )
                self.data["lambda"]["cw"]["logs"]["stream"] = span.attributes.pop(
                    "lambda.cw.logs.stream", None
                )
                self.data["lambda"]["cw"]["logs"]["more"] = span.attributes.pop(
                    "lambda.cw.logs.more", None
                )
                self.data["lambda"]["cw"]["logs"]["events"] = span.attributes.pop(
                    "lambda.cw.logs.events", None
                )

            elif trigger_type == "aws:s3":
                self.data["lambda"]["s3"]["events"] = span.attributes.pop(
                    "lambda.s3.events", None
                )
            elif trigger_type == "aws:sqs":
                self.data["lambda"]["sqs"]["messages"] = span.attributes.pop(
                    "lambda.sqs.messages", None
                )

        elif span.name == "celery-worker":
            self.data["celery"]["task"] = span.attributes.pop("task", None)
            self.data["celery"]["task_id"] = span.attributes.pop("task_id", None)
            self.data["celery"]["scheme"] = span.attributes.pop("scheme", None)
            self.data["celery"]["host"] = span.attributes.pop("host", None)
            self.data["celery"]["port"] = span.attributes.pop("port", None)
            self.data["celery"]["retry-reason"] = span.attributes.pop(
                "retry-reason", None
            )
            self.data["celery"]["error"] = span.attributes.pop("error", None)

        elif span.name == "gcps-consumer":
            self.data["gcps"]["op"] = span.attributes.pop("gcps.op", None)
            self.data["gcps"]["projid"] = span.attributes.pop("gcps.projid", None)
            self.data["gcps"]["sub"] = span.attributes.pop("gcps.sub", None)

        elif span.name == "rabbitmq":
            self.data["rabbitmq"]["exchange"] = span.attributes.pop("exchange", None)
            self.data["rabbitmq"]["queue"] = span.attributes.pop("queue", None)
            self.data["rabbitmq"]["sort"] = span.attributes.pop("sort", None)
            self.data["rabbitmq"]["address"] = span.attributes.pop("address", None)
            self.data["rabbitmq"]["key"] = span.attributes.pop("key", None)

        elif span.name == "rpc-server":
            self.data["rpc"]["flavor"] = span.attributes.pop("rpc.flavor", None)
            self.data["rpc"]["host"] = span.attributes.pop("rpc.host", None)
            self.data["rpc"]["port"] = span.attributes.pop("rpc.port", None)
            self.data["rpc"]["call"] = span.attributes.pop("rpc.call", None)
            self.data["rpc"]["call_type"] = span.attributes.pop("rpc.call_type", None)
            self.data["rpc"]["params"] = span.attributes.pop("rpc.params", None)
            # self.data["rpc"]["baggage"] = span.attributes.pop("rpc.baggage", None)
            self.data["rpc"]["error"] = span.attributes.pop("rpc.error", None)
        else:
            logger.debug("SpanRecorder: Unknown entry span: %s" % span.name)

    def _populate_local_span_data(self, span) -> None:
        if span.name == "render":
            self.data["render"]["name"] = span.attributes.pop("name", None)
            self.data["render"]["type"] = span.attributes.pop("type", None)
            self.data["log"]["message"] = span.attributes.pop("message", None)
            self.data["log"]["parameters"] = span.attributes.pop("parameters", None)
        else:
            logger.debug("SpanRecorder: Unknown local span: %s" % span.name)

    def _populate_exit_span_data(self, span) -> None:
        if span.name in HTTP_SPANS:
            self._collect_http_attributes(span)

        elif span.name == "boto3":
            # boto3 also sends http attributes
            self._collect_http_attributes(span)

            for attribute in ["op", "ep", "reg", "payload", "error"]:
                value = span.attributes.pop(attribute, None)
                if value is not None:
                    if attribute == "payload":
                        self.data["boto3"][attribute] = self._validate_attributes(value)
                    else:
                        self.data["boto3"][attribute] = value

        elif span.name == "cassandra":
            self.data["cassandra"]["cluster"] = span.attributes.pop(
                "cassandra.cluster", None
            )
            self.data["cassandra"]["query"] = span.attributes.pop(
                "cassandra.query", None
            )
            self.data["cassandra"]["keyspace"] = span.attributes.pop(
                "cassandra.keyspace", None
            )
            self.data["cassandra"]["fetchSize"] = span.attributes.pop(
                "cassandra.fetchSize", None
            )
            self.data["cassandra"]["achievedConsistency"] = span.attributes.pop(
                "cassandra.achievedConsistency", None
            )
            self.data["cassandra"]["triedHosts"] = span.attributes.pop(
                "cassandra.triedHosts", None
            )
            self.data["cassandra"]["fullyFetched"] = span.attributes.pop(
                "cassandra.fullyFetched", None
            )
            self.data["cassandra"]["error"] = span.attributes.pop(
                "cassandra.error", None
            )

        elif span.name == "celery-client":
            self.data["celery"]["task"] = span.attributes.pop("task", None)
            self.data["celery"]["task_id"] = span.attributes.pop("task_id", None)
            self.data["celery"]["scheme"] = span.attributes.pop("scheme", None)
            self.data["celery"]["host"] = span.attributes.pop("host", None)
            self.data["celery"]["port"] = span.attributes.pop("port", None)
            self.data["celery"]["error"] = span.attributes.pop("error", None)

        elif span.name == "couchbase":
            self.data["couchbase"]["hostname"] = span.attributes.pop(
                "couchbase.hostname", None
            )
            self.data["couchbase"]["bucket"] = span.attributes.pop(
                "couchbase.bucket", None
            )
            self.data["couchbase"]["type"] = span.attributes.pop("couchbase.type", None)
            self.data["couchbase"]["error"] = span.attributes.pop(
                "couchbase.error", None
            )
            self.data["couchbase"]["error_type"] = span.attributes.pop(
                "couchbase.error_type", None
            )
            self.data["couchbase"]["sql"] = span.attributes.pop("couchbase.sql", None)

        elif span.name == "rabbitmq":
            self.data["rabbitmq"]["exchange"] = span.attributes.pop("exchange", None)
            self.data["rabbitmq"]["queue"] = span.attributes.pop("queue", None)
            self.data["rabbitmq"]["sort"] = span.attributes.pop("sort", None)
            self.data["rabbitmq"]["address"] = span.attributes.pop("address", None)
            self.data["rabbitmq"]["key"] = span.attributes.pop("key", None)

        elif span.name == "redis":
            self.data["redis"]["connection"] = span.attributes.pop("connection", None)
            self.data["redis"]["driver"] = span.attributes.pop("driver", None)
            self.data["redis"]["command"] = span.attributes.pop("command", None)
            self.data["redis"]["error"] = span.attributes.pop("redis.error", None)
            self.data["redis"]["subCommands"] = span.attributes.pop("subCommands", None)

        elif span.name == "rpc-client":
            self.data["rpc"]["flavor"] = span.attributes.pop("rpc.flavor", None)
            self.data["rpc"]["host"] = span.attributes.pop("rpc.host", None)
            self.data["rpc"]["port"] = span.attributes.pop("rpc.port", None)
            self.data["rpc"]["call"] = span.attributes.pop("rpc.call", None)
            self.data["rpc"]["call_type"] = span.attributes.pop("rpc.call_type", None)
            self.data["rpc"]["params"] = span.attributes.pop("rpc.params", None)
            # self.data["rpc"]["baggage"] = span.attributes.pop("rpc.baggage", None)
            self.data["rpc"]["error"] = span.attributes.pop("rpc.error", None)

        elif span.name == "sqlalchemy":
            self.data["sqlalchemy"]["sql"] = span.attributes.pop("sqlalchemy.sql", None)
            self.data["sqlalchemy"]["eng"] = span.attributes.pop("sqlalchemy.eng", None)
            self.data["sqlalchemy"]["url"] = span.attributes.pop("sqlalchemy.url", None)
            self.data["sqlalchemy"]["err"] = span.attributes.pop("sqlalchemy.err", None)

        elif span.name == "mysql":
            self.data["mysql"]["host"] = span.attributes.pop("host", None)
            self.data["mysql"]["port"] = span.attributes.pop("port", None)
            self.data["mysql"]["db"] = span.attributes.pop(SpanAttributes.DB_NAME, None)
            self.data["mysql"]["user"] = span.attributes.pop(SpanAttributes.DB_USER, None)
            self.data["mysql"]["stmt"] = span.attributes.pop(SpanAttributes.DB_STATEMENT, None)
            self.data["mysql"]["error"] = span.attributes.pop("mysql.error", None)

        elif span.name == "postgres":
            self.data["pg"]["host"] = span.attributes.pop("host", None)
            self.data["pg"]["port"] = span.attributes.pop("port", None)
            self.data["pg"]["db"] = span.attributes.pop("db.name", None)
            self.data["pg"]["user"] = span.attributes.pop("db.user", None)
            self.data["pg"]["stmt"] = span.attributes.pop("db.statement", None)
            self.data["pg"]["error"] = span.attributes.pop("pg.error", None)

        elif span.name == "mongo":
            service = f"{span.attributes.pop(SpanAttributes.SERVER_ADDRESS, None)}:{span.attributes.pop(SpanAttributes.SERVER_PORT, None)}"
            namespace = f"{span.attributes.pop(SpanAttributes.DB_NAME, '?')}.{span.attributes.pop(SpanAttributes.DB_MONGODB_COLLECTION, '?')}"

            self.data["mongo"]["service"] = service
            self.data["mongo"]["namespace"] = namespace
            self.data["mongo"]["command"] = span.attributes.pop("command", None)
            self.data["mongo"]["filter"] = span.attributes.pop("filter", None)
            self.data["mongo"]["json"] = span.attributes.pop("json", None)
            self.data["mongo"]["error"] = span.attributes.pop("error", None)

        elif span.name == "gcs":
            self.data["gcs"]["op"] = span.attributes.pop("gcs.op", None)
            self.data["gcs"]["bucket"] = span.attributes.pop("gcs.bucket", None)
            self.data["gcs"]["object"] = span.attributes.pop("gcs.object", None)
            self.data["gcs"]["entity"] = span.attributes.pop("gcs.entity", None)
            self.data["gcs"]["range"] = span.attributes.pop("gcs.range", None)
            self.data["gcs"]["sourceBucket"] = span.attributes.pop(
                "gcs.sourceBucket", None
            )
            self.data["gcs"]["sourceObject"] = span.attributes.pop(
                "gcs.sourceObject", None
            )
            self.data["gcs"]["sourceObjects"] = span.attributes.pop(
                "gcs.sourceObjects", None
            )
            self.data["gcs"]["destinationBucket"] = span.attributes.pop(
                "gcs.destinationBucket", None
            )
            self.data["gcs"]["destinationObject"] = span.attributes.pop(
                "gcs.destinationObject", None
            )
            self.data["gcs"]["numberOfOperations"] = span.attributes.pop(
                "gcs.numberOfOperations", None
            )
            self.data["gcs"]["projectId"] = span.attributes.pop("gcs.projectId", None)
            self.data["gcs"]["accessId"] = span.attributes.pop("gcs.accessId", None)

        elif span.name == "gcps-producer":
            self.data["gcps"]["op"] = span.attributes.pop("gcps.op", None)
            self.data["gcps"]["projid"] = span.attributes.pop("gcps.projid", None)
            self.data["gcps"]["top"] = span.attributes.pop("gcps.top", None)

        elif span.name == "log":
            # use last special key values
            for event in span.events:
                if "message" in event.attributes:
                    self.data["log"]["message"] = event.attributes.pop("message", None)
                if "parameters" in event.attributes:
                    self.data["log"]["parameters"] = event.attributes.pop(
                        "parameters", None
                    )
        else:
            logger.debug("SpanRecorder: Unknown exit span: %s" % span.name)

    def _collect_http_attributes(self, span) -> None:
        self.data["http"]["host"] = span.attributes.pop("http.host", None)
        self.data["http"]["url"] = span.attributes.pop("http.url", None)
        self.data["http"]["path"] = span.attributes.pop("http.path", None)
        self.data["http"]["params"] = span.attributes.pop("http.params", None)
        self.data["http"]["method"] = span.attributes.pop("http.method", None)
        self.data["http"]["status"] = span.attributes.pop("http.status_code", None)
        self.data["http"]["path_tpl"] = span.attributes.pop("http.path_tpl", None)
        self.data["http"]["error"] = span.attributes.pop("http.error", None)

        if len(span.attributes) > 0:
            custom_headers = []
            for key in span.attributes:
                if key[0:12] == "http.header.":
                    custom_headers.append(key)

            for key in custom_headers:
                trimmed_key = key[12:]
                self.data["http"]["header"][trimmed_key] = span.attributes.pop(key)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

try:
    import tornado

    import wrapt
    import functools
    from typing import TYPE_CHECKING, Dict, Any

    from opentelemetry.semconv.trace import SpanAttributes

    from instana.log import logger
    from instana.singletons import agent, tracer
    from instana.util.secrets import strip_secrets_from_query
    from instana.util.traceutils import extract_custom_headers
    from instana.propagators.format import Format
    from instana.span.span import get_current_span


    @wrapt.patch_function_wrapper('tornado.httpclient', 'AsyncHTTPClient.fetch')
    def fetch_with_instana(wrapped, instance, argv, kwargs):
        try:
            parent_span = get_current_span()

            # If we're not tracing, just return
            if (not parent_span.is_recording()) or (parent_span.name == "tornado-client"):
                return wrapped(*argv, **kwargs)

            request = argv[0]

            # To modify request headers, we have to preemptively create an HTTPRequest object if a
            # URL string was passed.
            if not isinstance(request, tornado.httpclient.HTTPRequest):
                request = tornado.httpclient.HTTPRequest(url=request, **kwargs)

                new_kwargs = {}
                for param in ('callback', 'raise_error'):
                    # if not in instead and pop
                    if param in kwargs:
                        new_kwargs[param] = kwargs.pop(param)
                kwargs = new_kwargs

            parent_context = parent_span.get_span_context() if parent_span else None

            span = tracer.start_span("tornado-client", span_context=parent_context)

            extract_custom_headers(span, request.headers)

            tracer.inject(span.context, Format.HTTP_HEADERS, request.headers)

            # Query param scrubbing
            parts = request.url.split('?')
            if len(parts) > 1:
                cleaned_qp = strip_secrets_from_query(parts[1], agent.options.secrets_matcher,
                                                      agent.options.secrets_list)
                span.set_attribute("http.params", cleaned_qp)

            span.set_attribute(SpanAttributes.HTTP_URL, parts[0])
            span.set_attribute(SpanAttributes.HTTP_METHOD, request.method)

            future = wrapped(request, **kwargs)

            if future is not None:
                cb = functools.partial(finish_tracing, span=span)
                future.add_done_callback(cb)

            return future
        except Exception:
            logger.debug("Tornado fetch_with_instana: ", exc_info=True)


    def finish_tracing(future, span):
        try:
            response = future.result()
            span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, response.code)

            extract_custom_headers(span, response.headers)
        except tornado.httpclient.HTTPClientError as e:
            span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, e.code)
            span.record_exception(e)
            logger.debug("Tornado finish_tracing HTTPClientError: ", exc_info=True)
        finally:
            if span.is_recording():
                span.end()


    logger.debug("Instrumenting tornado client")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os

import urllib3
import pytest
from typing import Generator
from django.apps import apps
from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from instana.util.ids import hex_id
from tests.apps.app_django import INSTALLED_APPS
from instana.singletons import agent, tracer
from tests.helpers import (
    fail_with_message_and_span_dump,
    get_first_span_by_filter,
    drop_log_spans_from_list,
)
from instana.instrumentation.django.middleware import url_pattern_route

apps.populate(INSTALLED_APPS)


class TestDjango(StaticLiveServerTestCase):
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Setup and Teardown"""
        self.http = urllib3.PoolManager()
        self.recorder = tracer.span_processor
        # clear all spans before a test run
        self.recorder.clear_spans()
        yield
        # clear the INSTANA_DISABLE_W3C_TRACE_CORRELATION environment variable
        os.environ["INSTANA_DISABLE_W3C_TRACE_CORRELATION"] = ""

    def test_basic_request(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", self.live_server_url + "/", fields={"test": 1}
            )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 3 == len(spans)

        test_span = spans[2]
        urllib3_span = spans[1]
        django_span = spans[0]

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert "django" == django_span.n

        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == django_span.t

        assert urllib3_span.p == test_span.s
        assert django_span.p == urllib3_span.s

        assert django_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        assert django_span.ec is None
        assert "/" == django_span.data["http"]["url"]
        assert "GET" == django_span.data["http"]["method"]
        assert 200 == django_span.data["http"]["status"]
        assert "test=1" == django_span.data["http"]["params"]
        assert "^$" == django_span.data["http"]["path_tpl"]

        assert django_span.stack is None

    def test_synthetic_request(self) -> None:
        headers = {"X-INSTANA-SYNTHETIC": "1"}

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", self.live_server_url + "/", headers=headers
            )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 3 == len(spans)

        test_span = spans[2]
        urllib3_span = spans[1]
        django_span = spans[0]

        assert "^$" == django_span.data["http"]["path_tpl"]

        assert django_span.sy
        assert urllib3_span.sy is None
        assert test_span.sy is None

    def test_request_with_error(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", self.live_server_url + "/cause_error")

        assert response
        assert 500 == response.status

        spans = self.recorder.queued_spans()
        spans = drop_log_spans_from_list(spans)

        span_count = len(spans)
        if span_count != 3:
            msg = "Expected 3 spans but got %d" % span_count
            fail_with_message_and_span_dump(msg, spans)

        filter = lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "urllib3"
        urllib3_span = get_first_span_by_filter(spans, filter)
        assert urllib3_span

        filter = lambda span: span.n == "django"
        django_span = get_first_span_by_filter(spans, filter)
        assert django_span

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert "django" == django_span.n

        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == django_span.t

        assert urllib3_span.p == test_span.s
        assert django_span.p == urllib3_span.s

        assert 1 == django_span.ec

        assert "/cause_error" == django_span.data["http"]["url"]
        assert "GET" == django_span.data["http"]["method"]
        assert 500 == django_span.data["http"]["status"]
        assert "This is a fake error: /cause-error" == django_span.data["http"]["error"]
        assert "^cause_error$" == django_span.data["http"]["path_tpl"]
        assert django_span.stack is None

    def test_request_with_not_found(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", self.live_server_url + "/not_found")

        assert response
        assert 404 == response.status

        spans = self.recorder.queued_spans()
        spans = drop_log_spans_from_list(spans)

        span_count = len(spans)
        if span_count != 3:
            msg = "Expected 3 spans but got %d" % span_count
            fail_with_message_and_span_dump(msg, spans)

        filter = lambda span: span.n == "django"
        django_span = get_first_span_by_filter(spans, filter)
        assert django_span

        assert django_span.ec is None
        assert 404 == django_span.data["http"]["status"]

    def test_request_with_not_found_no_route(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", self.live_server_url + "/no_route")

        assert response
        assert 404 == response.status

        spans = self.recorder.queued_spans()
        spans = drop_log_spans_from_list(spans)

        span_count = len(spans)
        if span_count != 3:
            msg = "Expected 3 spans but got %d" % span_count
            fail_with_message_and_span_dump(msg, spans)

        filter = lambda span: span.n == "django"
        django_span = get_first_span_by_filter(spans, filter)
        assert django_span
        assert django_span.data["http"]["path_tpl"] is None
        assert django_span.ec is None
        assert 404 == django_span.data["http"]["status"]

    def test_complex_request(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", self.live_server_url + "/complex")

        assert response
        assert 200 == response.status
        spans = self.recorder.queued_spans()
        assert 5 == len(spans)

        test_span = spans[4]
        urllib3_span = spans[3]
        django_span = spans[2]
        otel_span1 = spans[1]
        otel_span2 = spans[0]

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert "django" == django_span.n
        assert "sdk" == otel_span1.n
        assert "sdk" == otel_span2.n

        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == django_span.t
        assert django_span.t == otel_span1.t
        assert otel_span1.t == otel_span2.t

        assert urllib3_span.p == test_span.s
        assert django_span.p == urllib3_span.s
        assert otel_span1.p == django_span.s
        assert otel_span2.p == otel_span1.s

        assert django_span.ec is None
        assert django_span.stack is None

        assert otel_span1.data["sdk"]["type"] == "exit"
        assert otel_span2.data["sdk"]["type"] == otel_span1.data["sdk"]["type"]
        otel_span1.data["sdk"]["name"] == "asteroid"
        otel_span2.data["sdk"]["name"] == "spacedust"

        assert "/complex" == django_span.data["http"]["url"]
        assert "GET" == django_span.data["http"]["method"]
        assert 200 == django_span.data["http"]["status"]
        assert "^complex$" == django_span.data["http"]["path_tpl"]

    def test_request_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        request_headers = {"X-Capture-This": "this", "X-Capture-That": "that"}

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", self.live_server_url + "/", headers=request_headers
            )
            # response = self.client.get('/')

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 3 == len(spans)

        test_span = spans[2]
        urllib3_span = spans[1]
        django_span = spans[0]

        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert "django" == django_span.n

        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == django_span.t

        assert urllib3_span.p == test_span.s
        assert django_span.p == urllib3_span.s

        assert django_span.ec is None
        assert django_span.stack is None

        assert "/" == django_span.data["http"]["url"]
        assert "GET" == django_span.data["http"]["method"]
        assert 200 == django_span.data["http"]["status"]
        assert "^$" == django_span.data["http"]["path_tpl"]

        assert "X-Capture-This" in django_span.data["http"]["header"]
        assert "this" == django_span.data["http"]["header"]["X-Capture-This"]
        assert "X-Capture-That" in django_span.data["http"]["header"]
        assert "that" == django_span.data["http"]["header"]["X-Capture-That"]

        agent.options.extra_http_headers = original_extra_http_headers

    def test_response_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", self.live_server_url + "/response_with_headers"
            )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 3 == len(spans)

        test_span = spans[2]
        urllib3_span = spans[1]
        django_span = spans[0]

        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert "django" == django_span.n

        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == django_span.t

        assert urllib3_span.p == test_span.s
        assert django_span.p == urllib3_span.s

        assert django_span.ec is None
        assert django_span.stack is None

        assert "/response_with_headers" == django_span.data["http"]["url"]
        assert "GET" == django_span.data["http"]["method"]
        assert 200 == django_span.data["http"]["status"]
        assert "^response_with_headers$" == django_span.data["http"]["path_tpl"]

        assert "X-Capture-This-Too" in django_span.data["http"]["header"]
        assert "this too" == django_span.data["http"]["header"]["X-Capture-This-Too"]
        assert "X-Capture-That-Too" in django_span.data["http"]["header"]
        assert "that too" == django_span.data["http"]["header"]["X-Capture-That-Too"]

        agent.options.extra_http_headers = original_extra_http_headers

    @pytest.mark.skip("Handled when type of trace and span ids are modified to str")
    def test_with_incoming_context(self) -> None:
        request_headers = dict()
        request_headers["X-INSTANA-T"] = "1"
        request_headers["X-INSTANA-S"] = "1"
        request_headers["traceparent"] = (
            "01-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01-788777"
        )
        request_headers["tracestate"] = (
            "rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE"
        )

        response = self.http.request(
            "GET", self.live_server_url + "/", headers=request_headers
        )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 1 == len(spans)

        django_span = spans[0]

        # assert django_span.t == '0000000000000001'
        # assert django_span.p == '0000000000000001'
        assert django_span.t == 1
        assert django_span.p == 1

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert "traceparent" in response.headers
        # The incoming traceparent header had version 01 (which does not exist at the time of writing), but since we
        # support version 00, we also need to pass down 00 for the version field.
        assert (
            "00-4bf92f3577b34da6a3ce929d0e0e4736-{}-01".format(django_span.s)
            == response.headers["traceparent"]
        )

        assert "tracestate" in response.headers
        assert (
            "in={};{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE".format(
                django_span.t, django_span.s
            )
            == response.headers["tracestate"]
        )

    @pytest.mark.skip("Handled when type of trace and span ids are modified to str")
    def test_with_incoming_context_and_correlation(self) -> None:
        request_headers = dict()
        request_headers["X-INSTANA-T"] = "1"
        request_headers["X-INSTANA-S"] = "1"
        request_headers["X-INSTANA-L"] = (
            "1, correlationType=web; correlationId=1234567890abcdef"
        )
        request_headers["traceparent"] = (
            "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"
        )
        request_headers["tracestate"] = (
            "rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE"
        )

        response = self.http.request(
            "GET", self.live_server_url + "/", headers=request_headers
        )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 1 == len(spans)

        django_span = spans[0]

        assert django_span.t == "a3ce929d0e0e4736"
        assert django_span.p == "00f067aa0ba902b7"
        assert django_span.ia.t == "a3ce929d0e0e4736"
        assert django_span.ia.p == "8357ccd9da194656"
        assert django_span.lt == "4bf92f3577b34da6a3ce929d0e0e4736"
        assert django_span.tp
        assert django_span.crtp == "web"
        assert django_span.crid == "1234567890abcdef"

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert "traceparent" in response.headers
        assert (
            "00-4bf92f3577b34da6a3ce929d0e0e4736-{}-01".format(django_span.s)
            == response.headers["traceparent"]
        )

        assert "tracestate" in response.headers
        assert (
            "in={};{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE".format(
                django_span.t, django_span.s
            )
            == response.headers["tracestate"]
        )

    @pytest.mark.skip("Handled when type of trace and span ids are modified to str")
    def test_with_incoming_traceparent_tracestate(self) -> None:
        request_headers = dict()
        request_headers["traceparent"] = (
            "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"
        )
        request_headers["tracestate"] = (
            "rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE"
        )

        response = self.http.request(
            "GET", self.live_server_url + "/", headers=request_headers
        )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 1 == len(spans)

        django_span = spans[0]

        assert (
            django_span.t == "a3ce929d0e0e4736"
        )  # last 16 chars from traceparent trace_id
        assert django_span.p == "00f067aa0ba902b7"
        assert django_span.ia.t == "a3ce929d0e0e4736"
        assert django_span.ia.p == "8357ccd9da194656"
        assert django_span.lt == "4bf92f3577b34da6a3ce929d0e0e4736"
        assert django_span.tp

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert "traceparent" in response.headers
        assert (
            "00-4bf92f3577b34da6a3ce929d0e0e4736-{}-01".format(django_span.s)
            == response.headers["traceparent"]
        )

        assert "tracestate" in response.headers
        assert (
            "in=a3ce929d0e0e4736;{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE".format(
                django_span.s
            )
            == response.headers["tracestate"]
        )

    @pytest.mark.skip("Handled when type of trace and span ids are modified to str")
    def test_with_incoming_traceparent_tracestate_disable_traceparent(self) -> None:
        os.environ["INSTANA_DISABLE_W3C_TRACE_CORRELATION"] = "1"
        request_headers = dict()
        request_headers["traceparent"] = (
            "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"
        )
        request_headers["tracestate"] = (
            "rojo=00f067aa0ba902b7,in=a3ce929d0e0e4736;8357ccd9da194656,congo=t61rcWkgMzE"
        )

        response = self.http.request(
            "GET", self.live_server_url + "/", headers=request_headers
        )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 1 == len(spans)

        django_span = spans[0]

        assert (
            django_span.t == "a3ce929d0e0e4736"
        )  # last 16 chars from traceparent trace_id
        assert django_span.p == "8357ccd9da194656"

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert "traceparent" in response.headers
        assert (
            "00-4bf92f3577b34da6a3ce929d0e0e4736-{}-01".format(django_span.s)
            == response.headers["traceparent"]
        )

        assert "tracestate" in response.headers
        assert (
            "in={};{},rojo=00f067aa0ba902b7,congo=t61rcWkgMzE".format(
                django_span.t, django_span.s
            )
            == response.headers["tracestate"]
        )

    def test_with_incoming_mixed_case_context(self) -> None:
        request_headers = dict()
        request_headers["X-InSTANa-T"] = "0000000000000001"
        request_headers["X-instana-S"] = "0000000000000001"

        response = self.http.request(
            "GET", self.live_server_url + "/", headers=request_headers
        )

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 1 == len(spans)

        django_span = spans[0]

        # assert django_span.t == '0000000000000001'
        # assert django_span.p == '0000000000000001'
        assert django_span.t == 1
        assert django_span.p == 1

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(django_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(django_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(django_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

    def test_url_pattern_route(self) -> None:
        view_name = "app_django.another"
        path_tpl = "".join(url_pattern_route(view_name))
        assert path_tpl == "^another$"

        view_name = "app_django.complex"
        try:
            path_tpl = "".join(url_pattern_route(view_name))
        except Exception:
            path_tpl = None
        assert path_tpl is None



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import threading
import time
from typing import Generator

import pytest

from instana.autoprofile.profiler import Profiler
from instana.autoprofile.runtime import RuntimeInfo
from instana.autoprofile.samplers.block_sampler import BlockSampler


class TestBlockSampler:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Create a new Profiler.
        self.profiler = Profiler(None)
        self.profiler.start(disable_timers=True)
        yield
        # teardown
        self.profiler.destroy()

    def test_block_profile(self) -> None:
        if RuntimeInfo.OS_WIN:
            return

        sampler = BlockSampler(self.profiler)
        sampler.setup()
        sampler.reset()

        lock = threading.Lock()
        event = threading.Event()

        def lock_lock() -> None:
            lock.acquire()
            time.sleep(0.5)
            lock.release()

        def lock_wait() -> None:
            lock.acquire()
            lock.release()

        def event_lock() -> None:
            time.sleep(0.5)
            event.set()

        def event_wait() -> None:
            event.wait()

        def record() -> None:
            sampler.start_sampler()
            time.sleep(2)
            sampler.stop_sampler()

        record_t = threading.Thread(target=record)
        record_t.start()

        # simulate lock
        t = threading.Thread(target=lock_lock)
        t.start()

        t = threading.Thread(target=lock_wait)
        t.start()

        # simulate event
        t = threading.Thread(target=event_lock)
        t.start()

        t = threading.Thread(target=event_wait)
        t.start()

        record_t.join()

        profile = sampler.build_profile(2000, 120000).to_dict()
        # print(profile)

        assert "lock_wait" in str(profile)
        assert "event_wait" in str(profile)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import copy
import wrapt

from typing import Callable, Optional, Any, Tuple, Dict
from instana.log import logger
from instana.instrumentation.pep0249 import ConnectionFactory

try:
    import psycopg2
    import psycopg2.extras

    cf = ConnectionFactory(connect_func=psycopg2.connect, module_name="postgres")

    setattr(psycopg2, "connect", cf)
    if hasattr(psycopg2, "Connect"):
        setattr(psycopg2, "Connect", cf)

    @wrapt.patch_function_wrapper("psycopg2.extensions", "register_type")
    def register_type_with_instana(
        wrapped: Callable[..., Any],
        instance: Optional[Any],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> Callable[..., object]:
        args_clone = list(copy.copy(args))

        if (len(args_clone) >= 2) and hasattr(args_clone[1], "__wrapped__"):
            args_clone[1] = args_clone[1].__wrapped__

        return wrapped(*args_clone, **kwargs)

    @wrapt.patch_function_wrapper("psycopg2._json", "register_json")
    def register_json_with_instana(
        wrapped: Callable[..., Any],
        instance: Optional[Any],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> Callable[..., object]:
        if "conn_or_curs" in kwargs:
            if hasattr(kwargs["conn_or_curs"], "__wrapped__"):
                kwargs["conn_or_curs"] = kwargs["conn_or_curs"].__wrapped__

        return wrapped(*args, **kwargs)

    logger.debug("Instrumenting psycopg2")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import gc
import logging
import os
import sys
import threading
from typing import Generator

import pytest
from instana.collector.helpers.runtime import (
    PATH_OF_AUTOTRACE_WEBHOOK_SITEDIR,
)
from instana.collector.host import HostCollector
from instana.singletons import get_agent, get_tracer
from instana.version import VERSION
from mock import patch
from pytest import LogCaptureFixture


class TestHostCollector:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.agent = get_agent()
        self.agent.collector = HostCollector(self.agent)
        self.tracer = get_tracer()
        self.webhook_sitedir_path = PATH_OF_AUTOTRACE_WEBHOOK_SITEDIR + "3.8.0"
        self.payload = None
        yield
        self.agent.collector.shutdown(report_final=False)
        variable_names = (
            "AWS_EXECUTION_ENV",
            "INSTANA_EXTRA_HTTP_HEADERS",
            "INSTANA_ENDPOINT_URL",
            "INSTANA_AGENT_KEY",
            "INSTANA_ZONE",
            "INSTANA_TAGS",
            "INSTANA_DISABLE_METRICS_COLLECTION",
            "INSTANA_DISABLE_PYTHON_PACKAGE_COLLECTION",
            "AUTOWRAPT_BOOTSTRAP",
        )

        for variable_name in variable_names:
            if variable_name in os.environ:
                os.environ.pop(variable_name)

        if self.webhook_sitedir_path in sys.path:
            sys.path.remove(self.webhook_sitedir_path)

    def test_start(self) -> None:
        with patch(
            "instana.collector.base.BaseCollector.is_reporting_thread_running",
            return_value=False,
        ):
            self.agent.collector.start()
            assert self.agent.collector.started
            assert self.agent.collector.THREAD_NAME == "Instana Collector"
            assert self.agent.collector.snapshot_data_interval == 300
            assert self.agent.collector.snapshot_data_last_sent == 0
            assert isinstance(self.agent.collector.helpers[0].collector, HostCollector)
            assert len(self.agent.collector.helpers) == 1
            assert isinstance(self.agent.collector.reporting_thread, threading.Thread)
            self.agent.collector.ready_to_start = False
            assert not self.agent.collector.start()

    def test_prepare_and_report_data(self, caplog: LogCaptureFixture) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        self.agent.collector.agent.machine.fsm.current = "wait4init"
        with patch("instana.agent.host.HostAgent.is_agent_ready", return_value=True):
            self.agent.collector.prepare_and_report_data()
            assert "Agent is ready.  Getting to work." in caplog.messages
            assert "Harmless state machine thread disagreement.  Will self-correct on next timer cycle."
        self.agent.collector.agent.machine.fsm.current = "wait4init"
        with patch("instana.agent.host.HostAgent.is_agent_ready", return_value=False):
            assert not self.agent.collector.prepare_and_report_data()
        self.agent.collector.agent.machine.fsm.current = "good2go"
        caplog.clear()
        with patch("instana.agent.host.HostAgent.is_timed_out", return_value=True):
            self.agent.collector.prepare_and_report_data()
            assert (
                "The Instana host agent has gone offline or is no longer reachable for > 1 min.  Will retry periodically."
                in caplog.messages
            )

    def test_should_send_snapshot_data(self) -> None:
        self.agent.collector.snapshot_data_interval = 999999999999
        assert not self.agent.collector.should_send_snapshot_data()

    def test_prepare_payload_basics(self) -> None:
        with patch.object(gc, "isenabled", return_value=True):
            self.payload = self.agent.collector.prepare_payload()
            assert self.payload

            assert len(self.payload.keys()) == 3
            assert "spans" in self.payload
            assert isinstance(self.payload["spans"], list)
            assert len(self.payload["spans"]) == 0
            assert "metrics", self.payload
            assert len(self.payload["metrics"].keys()) == 1
            assert "plugins", self.payload["metrics"]
            assert isinstance(self.payload["metrics"]["plugins"], list)
            assert len(self.payload["metrics"]["plugins"]) == 1

            python_plugin = self.payload["metrics"]["plugins"][0]
            assert python_plugin["name"] == "com.instana.plugin.python"
            assert python_plugin["entityId"] == str(os.getpid())
            assert "data" in python_plugin
            assert "snapshot" in python_plugin["data"]
            assert "m" in python_plugin["data"]["snapshot"]
            assert "Manual" == python_plugin["data"]["snapshot"]["m"]
            assert "metrics" in python_plugin["data"]

            assert "ru_utime" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_utime"]) in [float, int]
            assert "ru_stime" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_stime"]) in [float, int]
            assert "ru_maxrss" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_maxrss"]) in [float, int]
            assert "ru_ixrss" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_ixrss"]) in [float, int]
            assert "ru_idrss" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_idrss"]) in [float, int]
            assert "ru_isrss" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_isrss"]) in [float, int]
            assert "ru_minflt" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_minflt"]) in [float, int]
            assert "ru_majflt" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_majflt"]) in [float, int]
            assert "ru_nswap" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_nswap"]) in [float, int]
            assert "ru_inblock" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_inblock"]) in [float, int]
            assert "ru_oublock" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_oublock"]) in [float, int]
            assert "ru_msgsnd" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_msgsnd"]) in [float, int]
            assert "ru_msgrcv" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_msgrcv"]) in [float, int]
            assert "ru_nsignals" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_nsignals"]) in [float, int]
            assert "ru_nvcsw" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_nvcsw"]) in [float, int]
            assert "ru_nivcsw" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["ru_nivcsw"]) in [float, int]
            assert "alive_threads" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["alive_threads"]) in [
                float,
                int,
            ]
            assert "dummy_threads" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["dummy_threads"]) in [
                float,
                int,
            ]
            assert "daemon_threads" in python_plugin["data"]["metrics"]
            assert type(python_plugin["data"]["metrics"]["daemon_threads"]) in [
                float,
                int,
            ]

            assert "gc" in python_plugin["data"]["metrics"]
            assert isinstance(python_plugin["data"]["metrics"]["gc"], dict)
            assert "collect0" in python_plugin["data"]["metrics"]["gc"]
            assert type(python_plugin["data"]["metrics"]["gc"]["collect0"]) in [
                float,
                int,
            ]
            assert "collect1" in python_plugin["data"]["metrics"]["gc"]
            assert type(python_plugin["data"]["metrics"]["gc"]["collect1"]) in [
                float,
                int,
            ]
            assert "collect2" in python_plugin["data"]["metrics"]["gc"]
            assert type(python_plugin["data"]["metrics"]["gc"]["collect2"]) in [
                float,
                int,
            ]
            assert "threshold0" in python_plugin["data"]["metrics"]["gc"]
            assert type(python_plugin["data"]["metrics"]["gc"]["threshold0"]) in [
                float,
                int,
            ]
            assert "threshold1" in python_plugin["data"]["metrics"]["gc"]
            assert type(python_plugin["data"]["metrics"]["gc"]["threshold1"]) in [
                float,
                int,
            ]
            assert "threshold2" in python_plugin["data"]["metrics"]["gc"]
            assert type(python_plugin["data"]["metrics"]["gc"]["threshold2"]) in [
                float,
                int,
            ]

    def test_prepare_payload_basics_disable_runtime_metrics(self) -> None:
        os.environ["INSTANA_DISABLE_METRICS_COLLECTION"] = "TRUE"
        self.payload = self.agent.collector.prepare_payload()
        assert self.payload

        assert len(self.payload.keys()) == 3
        assert "spans" in self.payload
        assert isinstance(self.payload["spans"], list)
        assert len(self.payload["spans"]) == 0
        assert "metrics" in self.payload
        assert len(self.payload["metrics"].keys()) == 1
        assert "plugins" in self.payload["metrics"]
        assert isinstance(self.payload["metrics"]["plugins"], list)
        assert len(self.payload["metrics"]["plugins"]) == 1

        python_plugin = self.payload["metrics"]["plugins"][0]
        assert python_plugin["name"] == "com.instana.plugin.python"
        assert python_plugin["entityId"] == str(os.getpid())
        assert "data" in python_plugin
        assert "snapshot" in python_plugin["data"]
        assert "m" in python_plugin["data"]["snapshot"]
        assert "Manual" == python_plugin["data"]["snapshot"]["m"]
        assert "metrics" not in python_plugin["data"]

    def test_prepare_payload_with_snapshot_with_python_packages(self) -> None:
        self.payload = self.agent.collector.prepare_payload()
        assert self.payload
        assert "snapshot" in self.payload["metrics"]["plugins"][0]["data"]
        snapshot = self.payload["metrics"]["plugins"][0]["data"]["snapshot"]
        assert snapshot
        assert "m" in snapshot
        assert "Manual" == snapshot["m"]
        assert "version" in snapshot
        assert len(snapshot["versions"]) > 5
        assert snapshot["versions"]["instana"] == VERSION
        assert "wrapt" in snapshot["versions"]
        assert "fysom" in snapshot["versions"]

    def test_prepare_payload_with_snapshot_disabled_python_packages(self) -> None:
        os.environ["INSTANA_DISABLE_PYTHON_PACKAGE_COLLECTION"] = "TRUE"
        self.payload = self.agent.collector.prepare_payload()
        assert self.payload
        assert "snapshot" in self.payload["metrics"]["plugins"][0]["data"]
        snapshot = self.payload["metrics"]["plugins"][0]["data"]["snapshot"]
        assert snapshot
        assert "m" in snapshot
        assert "Manual" == snapshot["m"]
        assert "version" in snapshot
        assert len(snapshot["versions"]) == 1
        assert snapshot["versions"]["instana"] == VERSION

    def test_prepare_payload_with_autowrapt(self) -> None:
        os.environ["AUTOWRAPT_BOOTSTRAP"] = "instana"
        self.payload = self.agent.collector.prepare_payload()
        assert self.payload
        assert "snapshot" in self.payload["metrics"]["plugins"][0]["data"]
        snapshot = self.payload["metrics"]["plugins"][0]["data"]["snapshot"]
        assert snapshot
        assert "m" in snapshot
        assert "Autowrapt" == snapshot["m"]
        assert "version" in snapshot
        assert len(snapshot["versions"]) > 5
        expected_packages = ("instana", "wrapt", "fysom")
        for package in expected_packages:
            assert (
                package in snapshot["versions"]
            ), f"{package} not found in snapshot['versions']"
        assert snapshot["versions"]["instana"] == VERSION

    def test_prepare_payload_with_autotrace(self) -> None:
        sys.path.append(self.webhook_sitedir_path)
        self.payload = self.agent.collector.prepare_payload()
        assert self.payload
        assert "snapshot" in self.payload["metrics"]["plugins"][0]["data"]
        snapshot = self.payload["metrics"]["plugins"][0]["data"]["snapshot"]
        assert snapshot
        assert "m" in snapshot
        assert "AutoTrace" == snapshot["m"]
        assert "version" in snapshot
        assert len(snapshot["versions"]) > 5
        expected_packages = ("instana", "wrapt", "fysom")
        for package in expected_packages:
            assert (
                package in snapshot["versions"]
            ), f"{package} not found in snapshot['versions']"
        assert snapshot["versions"]["instana"] == VERSION
